#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading

from fim.user import GraphFormat

from fabric_cf.actor.core.apis.abc_actor_event import ABCActorEvent
from fabric_cf.actor.core.apis.abc_mgmt_controller_mixin import ABCMgmtControllerMixin
from fabric_cf.actor.core.apis.abc_tick import ABCTick
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.event_processor import EventProcessor
from fabric_cf.actor.core.manage.management_utils import ManagementUtils
from fabric_cf.actor.core.util.id import ID
from fabric_cf.orchestrator.core.bqm_wrapper import BqmWrapper
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.reservation_status_update_thread import ReservationStatusUpdateThread
from fabric_cf.orchestrator.core.slice_defer_thread import SliceDeferThread


class PollEvent(ABCActorEvent):
    def __init__(self, *, model_level_list: list):
        self.model_level_list = model_level_list

    def process(self):
        from fabric_cf.orchestrator.core.orchestrator_handler import OrchestratorHandler
        oh = OrchestratorHandler()
        for graph_format, level in self.model_level_list:
            oh.discover_broker_query_model(controller=oh.controller_state.controller,
                                           graph_format=graph_format, force_refresh=True, level=level)


class OrchestratorKernel(ABCTick):
    """
    Class responsible for starting Orchestrator Threads; also holds Management Actor and Broker information
    """

    def __init__(self):
        self.defer_thread = None
        self.sut = None
        self.broker = None
        self.logger = None
        self.controller = None
        self.lock = threading.Lock()
        self.bqm_cache = {}
        self.event_processor = None
        
    def get_saved_bqm(self, *, graph_format: GraphFormat, level: int) -> BqmWrapper:
        """
        Get Saved BQM from cache
        """
        try:
            self.lock.acquire()
            key = f"{graph_format}-{level}"
            saved_bqm = self.bqm_cache.get(key, None)
            return saved_bqm
        finally:
            self.lock.release()

    def save_bqm(self, *, bqm: str, graph_format: GraphFormat, level: int):
        try:
            self.lock.acquire()
            key = f"{graph_format}-{level}"
            saved_bqm = self.bqm_cache.get(key, None)
            if saved_bqm is None:
                from fabric_cf.actor.core.container.globals import GlobalsSingleton
                refresh_interval = GlobalsSingleton.get().get_config().get_global_config().get_bqm_config().get(
                    Constants.REFRESH_INTERVAL, None)
                saved_bqm = BqmWrapper()
                saved_bqm.set_refresh_interval(refresh_interval=int(refresh_interval))
            saved_bqm.save(bqm=bqm, graph_format=graph_format, level=level)
            self.bqm_cache[key] = saved_bqm
        finally:
            self.lock.release()

    def set_broker(self, *, broker: ID):
        """
        Set Broker
        :param broker:
        :return:
        """
        self.broker = broker

    def get_broker(self) -> ID:
        """
        Get Broker
        :return:
        """
        return self.broker

    def get_defer_thread(self) -> SliceDeferThread:
        return self.defer_thread

    def get_sut(self) -> ReservationStatusUpdateThread:
        """
        Get SUT thread
        :return:
        """
        return self.sut

    def get_logger(self):
        """
        Get logger
        :return:
        """
        if self.logger is None:
            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            self.logger = GlobalsSingleton.get().get_logger()
        return self.logger

    def get_management_actor(self) -> ABCMgmtControllerMixin:
        """
        Get Management actor
        :return:
        """
        if self.controller is None:
            self.controller = ManagementUtils.get_local_actor()
        return self.controller

    def stop_threads(self):
        """
        Stop threads
        :return:
        """
        if self.defer_thread is not None:
            self.defer_thread.stop()
        if self.event_processor is not None:
            self.event_processor.stop()
        #if self.sut is not None:
        #    self.sut.stop()

    def start_threads(self):
        """
        Start threads
        :return:
        """
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        GlobalsSingleton.get().get_container().register(tickable=self)

        self.get_logger().debug("Starting SliceDeferThread")
        self.defer_thread = SliceDeferThread(kernel=self)
        self.defer_thread.start()
        self.event_processor = EventProcessor(name="PeriodicProcessor", logger=self.logger)
        self.event_processor.start()
        #self.get_logger().debug("Starting ReservationStatusUpdateThread")
        #self.sut = ReservationStatusUpdateThread()
        #self.sut.start()

    def external_tick(self, *, cycle: int):
        """
        External tick
        @param cycle
        """
        try:
            self.lock.acquire()
            model_level_list = []
            for cached_bqm in self.bqm_cache.values():
                if cached_bqm.can_refresh():
                    model_level_list.append((cached_bqm.get_graph_format(), cached_bqm.get_level()))
            if self.event_processor is not None and len(model_level_list) > 0:
                self.event_processor.enqueue(incoming=PollEvent(model_level_list=model_level_list))
        except Exception as e:
            self.logger.error(f"Error occurred while doing periodic processing: {e}")
        finally:
            self.lock.release()

    def get_name(self) -> str:
        return self.__class__.__name__


class OrchestratorKernelSingleton:
    __instance = None

    def __init__(self):
        if self.__instance is not None:
            raise OrchestratorException("Singleton can't be created twice !")

    def get(self):
        """
        Actually create an instance
        """
        if self.__instance is None:
            self.__instance = OrchestratorKernel()
        return self.__instance

    get = classmethod(get)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import traceback

from fabric_cf.actor.core.apis.abc_client_reservation import ABCClientReservation

from fabric_cf.actor.core.policy.broker_simpler_units_policy import BrokerSimplerUnitsPolicy
from fabric_cf.actor.core.policy.controller_calendar_policy import ControllerCalendarPolicy
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.bids import Bids
from fabric_cf.actor.core.util.reservation_set import ReservationSet


class ControllerSimplePolicy(ControllerCalendarPolicy):
    """
    A simple implementation of a Controller policy. This version makes the following assumptions:
    - 1 broker
    - Always bid on only the first open auction
    - Extend all reservations expiring at bidding time, if renewable
    - Close - ADVANCE_CLOSE cycles early
    """
    # The amount of time over specific policy decisions the Controller must add when
    # communicating with other actors (e.g. redeem() and renew()). Clock skew
    # must be at least one if the Controller is ticked after the agent and/or authority
    # At some point in time we may want this to not be static and learn it from
    # what we see in the system, but for now it is static.
    CLOCK_SKEW = 1
    # How far in advance a reservation should initiate the close. This allows
    # for the Controller to property close its reservation before the authority does a
    # close on its behalf, which would eliminate any state the Controller needs to save.
    ADVANCE_CLOSE = 1

    def formulate_bids(self, *, cycle: int) -> Bids:
        """
        Form bids for expiring reservations and new demands. Return sets of reservations for new bids and renewals.
        formulateBids is unlocked on Controller. Note that a bidding policy never changes
        the state of reservations, except to suggest terms and brokers.
        It just returns sets of reservations ready to bid for and extend.
        It also never changes the membership of any reservation sets maintained by the server,
        although it does walk through them
        @params cycle: cycle
        @returns bids
        """
        extending = None
        bidding = None
        try:
            extending = self.process_renewing(cycle=cycle)
            # Select new reservations to bid, and bind to bid and term. Note:
            # here we issue all bids immediately. If we use a different policy,
            # it is our responsibility here to issue bids ahead of their
            # intended start cycles.
            bidding = self.process_demand(cycle=cycle)
            self.logger.debug("bidForSources: cycle {} bids {}".format(cycle, bidding.size()))
        except Exception as e:
            self.logger.error("an error in formulateBids:{}".format(e))
            self.logger.error(traceback.format_exc())

        return Bids(ticketing=bidding, extending=extending)

    def get_close(self, *, reservation: ABCClientReservation, term: Term) -> int:
        """
        Very simple policy - based on ADVANCE_CLOSE
        """
        if self.lazy_close:
            return -1
        else:
            end_cycle = self.actor.get_actor_clock().cycle(when=term.get_end_time())
            return end_cycle - self.ADVANCE_CLOSE

    def get_extend_term(self, *, suggested_term: Term, current_term: Term):
        """
        Returns the extension term for a reservation.
        @params suggested_term suggested term
        @params current_term current term
        @returns extension term
        @raises Exception in case of error
        """
        extend_term = None
        if suggested_term is not None:
            if suggested_term.extends_term(old_term=current_term):
                extend_term = suggested_term
            else:
                # extend the current term with the length of the term specified in suggested_term
                length = suggested_term.get_length()
                extend_term = current_term.extend(length=length)
        else:
            # Extend the term by its previous length
            extend_term = current_term.extend()
        return extend_term

    def get_redeem(self, *, reservation: ABCClientReservation) -> int:
        new_start = self.clock.cycle(when=reservation.get_term().get_new_start_time())
        result = new_start - self.CLOCK_SKEW
        if result < self.actor.get_current_cycle():
            result = self.actor.get_current_cycle()
        return result

    def get_renew(self, *, reservation: ABCClientReservation) -> int:
        """
        Call up to the agent to receive the advanceTime. Do time based on new_start so that requests are aligned.
        """
        new_start_cycle = self.actor.get_actor_clock().cycle(when=reservation.get_term().get_end_time()) + 1
        return new_start_cycle - BrokerSimplerUnitsPolicy.ADVANCE_TIME - self.CLOCK_SKEW

    def prepare(self, *, cycle: int):
        try:
            self.check_pending()
        except Exception as e:
            self.logger.error("Exception in prepare:{}".format(e))

    def process_demand(self, *, cycle: int) -> ReservationSet:
        """
        For each newly requested reservation, assigns a term to request, and a broker to bid from.
        @param cycle cycle
        @return non-null set of new bids
        @throws Exception in case of error rare
        """
        outgoing = ReservationSet()
        demand = self.calendar.get_demand()
        if demand is None:
            return ReservationSet()

        '''
        for reservation in demand.values():
            kernel_slice = reservation.get_slice()
            for slice_reservation in kernel_slice.get_reservations().values():
                self.logger.debug(f"Reservation {slice_reservation.get_reservation_id()} is in state: "
                                  f"{slice_reservation.get_state().name} type: {type(reservation)}")
        '''

        broker = self.actor.get_default_broker()
        for reservation in demand.values():
            if reservation.get_broker() is None:
                reservation.set_broker(broker=broker)

            rset = reservation.get_suggested_resources()
            term = reservation.get_suggested_term()
            reservation.set_approved(term=term, approved_resources=rset)
            outgoing.add(reservation=reservation)
            self.calendar.add_pending(reservation=reservation)
            self.calendar.remove_demand(reservation=reservation)
        return outgoing

    def process_renewing(self, *, cycle: int) -> ReservationSet:
        """
        Returns a fresh ReservationSet of expiring reservations to try to renew
        in this bidding cycle, and suggest new terms for them.
        @param cycle cycle
        @return non-null set of renewals
        @throws Exception in case of error rare
        """
        result = ReservationSet()
        renewing = self.calendar.get_renewing(cycle=cycle)
        if renewing is None or renewing.size() == 0:
            return result

        #self.logger.debug("Renewing = {}".format(renewing.size()))
        for reservation in renewing.values():
            #self.logger.debug("Renewing res: {}".format(reservation))

            if reservation.is_renewable():
                #self.logger.debug("Found a renewable reservation that needs an extension.")
                if reservation.is_closed() or reservation.is_closing() or reservation.is_failed():
                    self.logger.debug("Found a renewable reservation that is closing/closed/or failed")
                else:
                    suggested_term = reservation.get_suggested_term()
                    suggested_resources = reservation.get_suggested_resources()
                    current_term = reservation.get_term()
                    approved_resources = reservation.get_resources().abstract_clone()

                    approved_term = self.get_extend_term(suggested_term=suggested_term, current_term=current_term)
                    if suggested_resources is not None:
                        approved_resources.set_units(units=suggested_resources.get_units())
                        approved_resources.set_type(rtype=suggested_resources.get_type())
                        approved_resources.get_resource_data().merge(other=suggested_resources.get_resource_data())

                    reservation.set_approved(term=approved_term, approved_resources=approved_resources)
                    result.add(reservation=reservation)
                    self.calendar.add_pending(reservation=reservation)
            #else:
                #self.logger.error("A non-renewable reservation is on the renewing list")
        return result



from connexion.apps.flask_app import FlaskJSONEncoder
import six

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model


class JSONEncoder(FlaskJSONEncoder):
    include_nulls = False

    def default(self, o):
        if isinstance(o, Model):
            dikt = {}
            for attr, _ in six.iteritems(o.swagger_types):
                value = getattr(o, attr)
                if value is None and not self.include_nulls:
                    continue
                attr = o.attribute_map[attr]
                dikt[attr] = value
            return dikt
        return FlaskJSONEncoder.default(self, o)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
from datetime import datetime

from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.time.calendar.base_calendar import BaseCalendar
from fabric_cf.actor.core.util.reservation_holdings import ReservationHoldings
from fabric_cf.actor.core.util.reservation_list import ReservationList
from fabric_cf.actor.core.util.reservation_set import ReservationSet
from fabric_cf.actor.core.util.resource_type import ResourceType


class ClientCalendar(BaseCalendar):
    """
    This a client-side calendar to be used by brokers  or service managers. A client
    calendar maintains the following lists:
    - demand: a list of reservations representing demand for resources
    - pending: a list of reservations with pending operations
    - renewing: a list of reservations organized by the time they must be renewed (cycle)
    - holdings: a list of active/granted reservations associated with their lease term.
      This list should be maintained using real time, not cycles.

    The renewing and holding lists are automatically purged by the implementation as
    time advances. The demand and pending list, however, must be purged manually by the
    user of this class.
    """
    def __init__(self, *, clock: ActorClock):
        """
        Constructor
        @params clock: clock factory
        """
        super().__init__(clock=clock)
        # Set of reservations representing the current demand. Callers are
        # responsible for removing serviced reservations
        self.demand = ReservationSet()
        # Set of reservations for which a request has been issued but no
        # confirmation has been received. Callers are responsible for removing
        # acknowledged reservations.
        self.pending = ReservationSet()
        # Set of reservations grouped by renewing time.
        self.renewing = ReservationList()
        # Set of active reservations.
        self.holdings = ReservationHoldings()
        self.lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['lock']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.lock = threading.Lock()

    def remove(self, *, reservation: ABCReservationMixin):
        """
        Removes the specified reservation from all internal calendar data structures
        @params reservation: reservation to remove
        """
        self.remove_demand(reservation=reservation)
        self.remove_pending(reservation=reservation)
        self.remove_renewing(reservation=reservation)
        self.remove_holdings(reservation=reservation)

    def remove_scheduled_or_in_progress(self, *, reservation: ABCReservationMixin):
        """
        Removes the specified reservations from all internal calendar data
        structures that represent operations to be scheduled in the future or
        operations that are currently in progress. Does not remove the
        reservation from the holdings list.
        @params reservation : reservation to remove
        """
        self.remove_demand(reservation=reservation)
        self.remove_pending(reservation=reservation)
        self.remove_renewing(reservation=reservation)

    def get_demand(self) -> ReservationSet:
        """
        Returns the known demand. Can be for resources starting at different times.
        @returns the set of demanded reservations.
        """
        try:
            self.lock.acquire()
            return self.demand.clone()
        finally:
            self.lock.release()

    def add_demand(self, *, reservation: ABCReservationMixin):
        """
        Adds a reservation to the demand list.
        @params reservation: reservation to add
        """
        try:
            self.lock.acquire()
            self.demand.add(reservation=reservation)
        finally:
            self.lock.release()

    def remove_demand(self, *, reservation: ABCReservationMixin):
        """
        Removes a reservation to the demand list.
        @params reservation: reservation to remove
        """
        try:
            self.lock.acquire()
            self.demand.remove(reservation=reservation)
        finally:
            self.lock.release()

    def get_pending(self) -> ReservationSet:
        """
        Returns the known pending. Can be for resources starting at different times.
        @returns the set of pending reservations.
        """
        try:
            self.lock.acquire()
            return self.pending.clone()
        finally:
            self.lock.release()

    def add_pending(self, *, reservation: ABCReservationMixin):
        """
        Adds a reservation to the pending list.
        @params reservation: reservation to add
        """
        try:
            self.lock.acquire()
            self.pending.add(reservation=reservation)
        finally:
            self.lock.release()

    def remove_pending(self, *, reservation: ABCReservationMixin):
        """
        Removes a reservation to the pending list.
        @params reservation: reservation to remove
        """
        try:
            self.lock.acquire()
            self.pending.remove(reservation=reservation)
        finally:
            self.lock.release()

    def get_renewing(self, *, cycle: int) -> ReservationSet:
        """
        Returns the reservations that need to be renewed on the specified cycle.
        @params cycle : cycle number
        @returns reservation set with reservations to be renewed on the specified cycle
        """
        try:
            self.lock.acquire()
            return self.renewing.get_all_reservations(cycle=cycle)
        finally:
            self.lock.release()

    def add_renewing(self, *, reservation: ABCReservationMixin, cycle: int):
        """
        Adds a reservation to the renewing list at the given cycle.
        @params reservation : reservation to add
        @params cycle : cycle number
        """
        try:
            self.lock.acquire()
            self.renewing.add_reservation(reservation=reservation, cycle=cycle)
        finally:
            self.lock.release()

    def remove_renewing(self, *, reservation: ABCReservationMixin):
        """
        Removes the reservation from the renewing list.
        @params reservation : reservation to remove
        """
        try:
            self.lock.acquire()
            self.renewing.remove_reservation(reservation=reservation)
        finally:
            self.lock.release()

    def get_holdings(self, *, d: datetime = None, type: ResourceType = None) -> ReservationSet:
        """
        Returns the resources of the specified type held by the client that are active at the specified time instance.
        @params d : datetime instance.
        @params type : resource type
        @returns st of reservations of the specified type that are active at the specified time
        """
        try:
            self.lock.acquire()
            when = None
            if d is not None:
                when = ActorClock.to_milliseconds(when=d)
            return self.holdings.get_reservations(time=when, rtype=type)
        finally:
            self.lock.release()

    def add_holdings(self, *, reservation: ABCReservationMixin, start: datetime, end: datetime):
        """
        Adds a reservation to the holdings list.
        @params reservation : reservation to add
        @params start : start time
        @params end : end time
        """
        try:
            self.lock.acquire()
            self.holdings.add_reservation(reservation=reservation, start=ActorClock.to_milliseconds(when=start),
                                          end=ActorClock.to_milliseconds(when=end))
        finally:
            self.lock.release()

    def remove_holdings(self, *, reservation: ABCReservationMixin):
        """
        Removes the reservation from the renewing list.
        @params reservation : reservation to remove
        """
        try:
            self.lock.acquire()
            self.holdings.remove_reservation(reservation=reservation)
        finally:
            self.lock.release()

    def tick(self, *, cycle: int):
        try:
            self.lock.acquire()
            super().tick(cycle=cycle)
            self.renewing.tick(cycle=cycle)
            ms = self.clock.cycle_end_in_millis(cycle=cycle)
            self.holdings.tick(time=ms)
        finally:
            self.lock.release()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from enum import Enum
from typing import Tuple

from fabric_mb.message_bus.messages.reservation_mng import ReservationMng

from fabric_cf.actor.core.apis.abc_mgmt_controller_mixin import ABCMgmtControllerMixin
from fabric_cf.actor.core.util.id import ID


class Status(Enum):
    OK = 0
    NOT_OK = 1
    NOT_READY = 3


class StatusChecker:
    def __init__(self):
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()

    def check(self, *, controller: ABCMgmtControllerMixin, rid: ID) -> Tuple[Status, ReservationMng or None]:
        """
        Check status of the reservation identified by rid
        :param controller: controller
        :param rid: reservation id
        :return: Status
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, List, Dict


from fabric_cf.actor.boot.configuration import ActorConfig
from fabric_cf.actor.core.apis.abc_actor_runnable import ABCActorRunnable
from fabric_cf.actor.core.apis.abc_timer_queue import ABCTimerQueue
from fabric_cf.actor.core.apis.abc_actor_identity import ABCActorIdentity

from fabric_cf.actor.core.apis.abc_tick import ABCTick

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_event import ABCActorEvent
    from fabric_cf.actor.core.apis.abc_actor_proxy import ABCActorProxy
    from fabric_cf.actor.core.apis.abc_base_plugin import ABCBasePlugin
    from fabric_cf.actor.core.apis.abc_query_response_handler import ABCQueryResponseHandler
    from fabric_cf.actor.core.kernel.failed_rpc import FailedRPC
    from fabric_cf.actor.core.time.actor_clock import ActorClock
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.security.auth_token import AuthToken
    from fabric_cf.actor.core.apis.abc_policy import ABCPolicy
    from fabric_cf.actor.core.apis.abc_slice import ABCSlice
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
    from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
    from fabric_cf.actor.core.util.reservation_set import ReservationSet
    from fabric_cf.actor.core.kernel.resource_set import ResourceSet
    from fabric_cf.actor.core.time.term import Term
    from fabric_cf.actor.core.kernel.poa import Poa
    from fabric_cf.actor.core.container.maintenance import Site


class ActorType(Enum):
    """
    Enum for Actor Type
    """
    All = 0
    Orchestrator = 1
    Broker = 2
    Authority = 3

    @staticmethod
    def get_actor_type_from_string(*, actor_type: str) -> ActorType:
        """
        Convert String to Actor Type
        @param actor_type actor type
        """
        if actor_type.lower() == ActorType.Orchestrator.name.lower():
            return ActorType.Orchestrator
        if actor_type.lower() == ActorType.Broker.name.lower():
            return ActorType.Broker
        if actor_type.lower() == ActorType.Authority.name.lower():
            return ActorType.Authority
        return ActorType.All


class ABCActorMixin(ABCActorIdentity, ABCTick, ABCTimerQueue):
    """
     IActor defines the common functionality of all actors. An actor
     offers a collection of management operations for slices and reservations and
     implements the public methods necessary to serve calls from other actors,
     e.g., requests for tickets and leases.

     Every actor has a globally unique identifier and a name. The current
     implementation assumes that names are globally unique. In addition, each
     actor can have an optional description (used for display purposes, e.g., the
     web portal).

     The actions of each actor, e.g, how to request new resources, how to
     arbitrate among multiple ticket requests, etc, are driven by policy modules.

     There are three types of actors:
         Orchestrator -
         Broker - arbiter among requests for resources. Brokers determine who gets what and for how long.
         Aggregate Manager - owner of resources.

         Each of the aforementioned roles is defined in a corresponding interface. An
         actor instance must implement at least one of these interfaces.
    """

    @abstractmethod
    def actor_added(self, *, config: ActorConfig):
        """
        Informs the actor that it has been integrated in the container. This
        method should finish the initialization of the actor: some initialization
        steps may not be able to execute until the actor is part of the running
        container.

        Raises:
            Exception: if a critical error occurs while processing the event
        """

    @abstractmethod
    def actor_removed(self):
        """
        Informs the actor that it has been removed. This method should finish the
        shutdown/cleanup of the actor.
        """

    @abstractmethod
    def get_actor_clock(self) -> ActorClock:
        """
        Returns the actor clock used by the actor.

        Returns:
            actor clock
        """

    @abstractmethod
    def get_current_cycle(self) -> int:
        """
        Returns the cycle this actor is processing.

        Returns:
            current clock cycle
        """

    @abstractmethod
    def get_description(self) -> str:
        """
        Returns the description for the actor.

        Returns:
            description for the actor.
        """

    @abstractmethod
    def get_policy(self) -> ABCPolicy:
        """
        Returns the policy used by the actor.

        Returns:
            policy used by the actor.
        """

    @abstractmethod
    def get_plugin(self) -> ABCBasePlugin:
        """
        Returns the plugin used by the actor.

        Returns:
            plugin used by the actor.
        """

    @abstractmethod
    def get_type(self) -> ActorType:
        """
        Returns the type of the actor.

        Returns:
            type of the actor.
        """

    @abstractmethod
    def initialize(self, *, config: ActorConfig):
        """
        Initializes the actor.

        Raises:
            Exception: if a critical error occurs while initialization
        """

    @abstractmethod
    def is_recovered(self) -> bool:
        """
        Checks if the actor has completed recovery.

        Returns:
            true if this actor has completed recovery
        """

    @abstractmethod
    def is_stopped(self) -> bool:
        """
        Checks if the actor has completed stopped.

        Returns:
            true if this actor has been stopped
        """

    @abstractmethod
    def recover(self):
        """
        Recovers the actor from saved state.

        Raises:
            Exception: if an error occurs during recovery
        """

    @abstractmethod
    def set_actor_clock(self, *, clock: ActorClock):
        """
        Sets the actor clock to be used by the actor.

        Args:
            clock: actor clock
        """

    @abstractmethod
    def set_description(self, *, description: str):
        """
        Sets the description for the actor.

        Args:
            description: actor description
        """

    @abstractmethod
    def set_identity(self, *, token: AuthToken):
        """
        Sets the identity of this actor. Must be called before initialize.

        Args:
            token: actor's identity token
        """

    @abstractmethod
    def set_policy(self, *, policy: ABCPolicy):
        """
        Sets the policy of this actor. Must be called before initialize.

        Args:
            policy: policy implementation to use
        """

    @abstractmethod
    def set_recovered(self, *, value: bool):
        """
        Sets the recovered flag.

        Args:
            value: flag value
        """

    @abstractmethod
    def set_plugin(self, *, plugin: ABCBasePlugin):
        """
        Sets the plugin of this actor. Must be called before initialize.

        Args:
            plugin: plugin to use.
        """

    @abstractmethod
    def start(self):
        """
        Performs all required actions when starting an actor.
        """

    @abstractmethod
    def stop(self):
        """
        Performs all required actions when stopping an actor.
        """

    @abstractmethod
    def queue_event(self, *, incoming: ABCActorEvent):
        """
        Adds an event.

        Args:
            incoming: incoming event
        """

    @abstractmethod
    def query(self, *, query: dict = None, caller: AuthToken = None,
              actor_proxy: ABCActorProxy = None, handler: ABCQueryResponseHandler = None):
        """
        Processes a query request from the specified caller.

        Args:
            query: query
            caller: caller
            actor_proxy: actor proxy
            handler: handler

        Returns:
            query response
        """
    @abstractmethod
    def execute_on_actor_thread(self, *, runnable: ABCActorRunnable):
        """
        Execute on Actor Thread and Wait until response is processed
        @params runnable: reservation to be processed
        """

    @abstractmethod
    def execute_on_actor_thread_and_wait(self, *, runnable: ABCActorRunnable):
        """
        Execute on Actor Thread and Wait until response is processed
        @params runnable: reservation to be processed
        """

    @abstractmethod
    def await_no_pending_reservations(self):
        """
        Await for pending reservations
        """

    @abstractmethod
    def get_logger(self):
        """
        Return the logger
        """

    @abstractmethod
    def handle_failed_rpc(self, *, rid: ID, rpc: FailedRPC):
        """
        Handle a failed rpc
        @params rid: reservation id
        @params rpc: failed rpc
        """

    @staticmethod
    def get_management_object_class() -> str:
        """
        Get Management Object class Name
        """

    @staticmethod
    def get_management_object_module() -> str:
        """
        Get Management Object class Module name
        """

    @staticmethod
    def get_kafka_service_class() -> str:
        """
        Get Kafka Service Class Name
        """

    @staticmethod
    def get_kafka_service_module() -> str:
        """
        Get Kafka Service Class Module Name
        """

    @staticmethod
    def get_mgmt_kafka_service_class() -> str:
        """
        Get Management Kafka Service Class Name
        """

    @staticmethod
    def get_mgmt_kafka_service_module() -> str:
        """
        Get Management Kafka Service Class Module Name
        """

    @abstractmethod
    def set_logger(self, logger):
        """
        Set logger
        @param logger logger
        """

    @abstractmethod
    def load_model(self, *, graph_id: str):
        """
        Load any Graph Model
        :param graph_id:
        :return:
        """

    @abstractmethod
    def get_client_slices(self) -> List[ABCSlice]:
        """
        Returns all client slices registered with the actor.

        Returns:
            an array of client slices
        """

    @abstractmethod
    def get_slices(self) -> List[ABCSlice]:
        """
        Returns all slices registered with the actor.

        Returns:
            an array of slices
        """

    @abstractmethod
    def get_slice(self, *, slice_id: ID) -> ABCSlice:
        """
        Returns the slice with the given id.

        Args:
            slice_id: slice id
        Returns:
            the slice
        """

    @abstractmethod
    def register_slice(self, *, slice_object: ABCSlice):
        """
        Registers the slice with the actor. The slice must be a newly created one without a database record.
        If the slice is a recovered/previously unregistered one use re_register_slice(Slice) instead.

        Args:
            slice_object: slice
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def modify_slice(self, *, slice_object: ABCSlice):
        """
        Modify the slice registered with the actor. Moves the slice into Modifying State

        Args:
            slice_object: slice_object
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def delete_slice(self, *, slice_id: ID):
        """
        Delete the slice registered with the actor. Moves the slice into Closing State

        Args:
            slice_id: slice id
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def remove_slice(self, *, slice_object: ABCSlice):
        """
        Removes the specified slice. Purges slice-related state from the database.

        Args:
            slice_object: slice
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def remove_slice_by_slice_id(self, *, slice_id: ID):
        """
        Removes the specified slice. Purges slice-related state from the database.

        Args:
            slice_id: slice id
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def modify_accept(self, *, slice_id: ID):
        """
        Accept the last modify

        Args:
            slice_id: slice id
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def re_register_slice(self, *, slice_object: ABCSlice):
        """
        Re-registers the slice with the actor. The slice must already have a database record.

        Args:
            slice_object: slice
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def unregister_slice(self, *, slice_object: ABCSlice):
        """
        Unregisters the slice. Does not purge slice-related state from the database.

        Args:
            slice_object: slice
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def unregister_slice_by_slice_id(self, *, slice_id: ID):
        """
        Unregisters the slice. Does not purge slice-related state from the database.

        Args:
            slice_id: slice_id
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def close_slice_reservations(self, *, slice_id: ID):
        """
        Close slice reservations

        Args:
            slice_id: slice_id
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def fail(self, *, rid: ID, message: str):
        """
        Fails the specified reservation.

        Args:
            rid: reservation id
            message: message
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def fail_delegation(self, *, did: str, message: str):
        """
        Fails the specified delegation.

        Args:
            did: delegation id
            message: message
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def close_by_rid(self, *, rid: ID, force: bool = False):
        """
        Closes the reservation. Note: the reservation must have already been registered with the actor.
        This method may involve either a client or a server side action or both. When called on a broker,
        this method will only close the broker reservation.

        Args:
            rid: reservation id
            force: force close
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def close_delegation(self, *, did: str):
        """
        Closes the delegation. Note: the delegation must have already been registered with the actor.
        This method may involve either a client or a server side action or both. When called on a broker,
        this method will only close the broker delegation.

        Args:
            did: delegation id
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def close(self, *, reservation: ABCReservationMixin):
        """
        Closes the reservation. Note: the reservation must have already been registered with the actor.
        This method may involve either a client or a server side action or both. When called on a broker,
        this method will only close the broker reservation.

        Args:
            reservation: reservation
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def close_reservations(self, *, reservations: ReservationSet):
        """
        Issues a close for every reservation in the set. All exceptions are caught and logged but no exception
        is propagated. No information will be delivered to indicate that some failure has taken place, e.g.,
        failure to communicate with a broker. Inspect the state of individual reservations to
        determine whether/what failures have taken place.

        Args:
            reservations: reservation set
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def extend(self, *, rid: ID, resources: ResourceSet, term: Term, dependencies: List[ABCReservationMixin] = None):
        """
        Extends the reservation. Note: the reservation must have already been registered with the actor.
        This method may involve either a client or a server side action or both. When called on a broker,
        this method will extend the current ticket and send an update back to the client. When called on a
        site authority, this method will extend the lease and send the new lease back.

        Args:
            reservation: reservation to extend
            rid: reservation id
            resources: resource set describing the resources desired for the extension
            term: term for extension (must extend the current term)
            dependencies: dependencies
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def get_reservation(self, *, rid: ID) -> ABCReservationMixin:
        """
        Returns the specified reservation.

        Args:
            rid: reservation id
        Returns:
            reservation
        """

    @abstractmethod
    def get_reservations(self, *, slice_id: ID) -> List[ABCReservationMixin]:
        """
        Returns all reservations in the given slice.

        Args:
            slice_id: slice id
        Returns:
            reservations
        """

    @abstractmethod
    def register(self, *, reservation: ABCReservationMixin):
        """
        Registers the reservation with the actor. The reservation must not have been previously
        registered with the actor: there should be no database record for the reservation.

        Args:
            reservation: reservation to register
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def register_delegation(self, *, delegation: ABCDelegation):
        """
        Registers the delegation with the actor. The delegation must not have been previously
        registered with the actor: there should be no database record for the delegation.

        Args:
            delegation: delegation to register
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def remove_reservation(self, *, reservation: ABCReservationMixin = None, rid: ID = None):
        """
        Removes the specified reservation. Note: the reservation must have already been registered with the actor.
        This method will unregister the reservation and remove it from the underlying database.
        Only closed and failed reservations can be removed.

        Args:
            reservation: reservation
            rid: reservation id
        Raises:
            Exception if an error occurs or when trying to remove a reservation that is neither failed or closed.
        """

    @abstractmethod
    def re_register(self, *, reservation: ABCReservationMixin):
        """
        Registers a previously registered reservation with the actor. The reservation must have a database record.

        Args:
            reservation: reservation
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def re_register_delegation(self, *, delegation: ABCDelegation):
        """
        Registers a previously registered delegation with the actor. The delegation must have a database record.

        Args:
            delegation: delegation
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def unregister(self, *, reservation: ABCReservationMixin, rid: ID):
        """
        Unregisters the reservation with the actor. The reservation's database record will not be removed.

        Args:
            reservation: reservation
            rid: reservation id
        Raises:
            Exception in case of error
        """

    def get_asm_thread(self):
        return None

    @abstractmethod
    def remove_delegation(self, *, did: str):
        """
        Removes the specified delegation. Note: the delegation must have already been registered with the actor.
        This method will unregister the reservation and remove it from the underlying database.
        Only closed and failed delegation can be removed.

        Args:
            did: delegation id
        Raises:
            Exception if an error occurs or when trying to remove a delegation that is neither failed or closed.
        """

    @abstractmethod
    def update_maintenance_mode(self, *, properties: Dict[str, str], sites: List[Site] = None):
        """
        Update Maintenance mode
        @param properties properties
        @param sites sites

        @raises Exception in case of failure
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.util.reservation_holdings import ReservationHoldings
from fabric_cf.actor.core.util.reservation_list import ReservationList


class SourceCalendar:
    """
    SourceCalendar organizes state for a delegation used as a
    source for client reservations. A source calendar maintains a list of
    "outlays", client reservations that have been allocated from the source
    delegation. The outlay calendar is organized by real time.
    The calendar also maintains a list of incoming extension requests for
    reservations that have been satisfied from the underlying source delegations.
    """
    def __init__(self, *, clock: ActorClock, source: ABCDelegation):
        """
        Constructor
        @params clock: clock
        @params source: source delegation
        """
        # Clock.
        self.clock = clock
        # The source delegation.
        self.source = source
        # Allocated reservations.
        self.outlays = ReservationHoldings()
        # Incoming extension requests.
        self.extending = ReservationList()

    def tick(self, *, cycle: int):
        ms = self.clock.cycle_end_in_millis(cycle=cycle)
        self.outlays.tick(time=ms)
        self.extending.tick(cycle=cycle)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_container_clock import ABCContainerClock

if TYPE_CHECKING:
    from fabric_cf.actor.boot.configuration import Configuration
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.container.protocol_descriptor import ProtocolDescriptor
    from fabric_cf.actor.core.manage.management_object_manager import ManagementObjectManager
    from fabric_cf.actor.core.util.id import ID


class ABCActorContainer(ABCContainerClock):
    """
    IActorContainer is the public interface for container
    """
    @abstractmethod
    def initialize(self, *, config: Configuration):
        """
        Initializes the container manager.
        @param config container configuration
        @throws ContainerInitializationException if the configuration is invalid
        """

    @abstractmethod
    def get_guid(self) -> ID:
        """
        Returns the container GUID.
        @return container GUID
        """

    @abstractmethod
    def get_config(self) -> Configuration:
        """
        Return the container configuration
        @return container configuration
        """

    @abstractmethod
    def get_database(self):
        """
        Returns the container database.
        @return container database
        """

    @abstractmethod
    def is_recovered(self):
        """
        Checks if the container has completed recovery.
        @return TRUE if recovery is complete, FALSE otherwise
        """

    @abstractmethod
    def shutdown(self):
        """
        Shuts down the container.
        """

    @abstractmethod
    def register_actor(self, *, actor: ABCActorMixin):
        """
        Registers a new actor: adds the actor to the database, deploys services
        required by the actor, registers actor proxies and callbacks. Must not
        register the actor with the clock! Clock registration is a separate
        phase.
        @param actor actor to register
        @throws Exception in case of error
        """

    @abstractmethod
    def register_protocol(self, *, protocol: ProtocolDescriptor):
        """
        Registers a communication protocol with the container. This protocol
        applies only for internal communication among actors; this is not a
        protocol used for managing the container.
        @param protocol protocol to register
        """

    @abstractmethod
    def get_management_object_manager(self) -> ManagementObjectManager:
        """
        Returns Management Object Manager
        @returns ManagementObjectManager
        """

    @abstractmethod
    def get_protocol_descriptor(self, *, protocol: str) -> ProtocolDescriptor:
        """
        Return Protocol Descriptor for a specific Protocol
        @returns ProtocolDescriptor
        """

    @abstractmethod
    def is_fresh(self) -> bool:
        """
        Returns true if container was started fresh; false otherwise
        @returns bool
        """

    @abstractmethod
    def recover_actor(self, *, properties: dict):
        """
        Recover an actor
        @params properties: properties
        """

    @abstractmethod
    def unregister_actor(self, *, actor: ABCActorMixin):
        """
        Un-register an actor
        @params actor: actor
        """

    @abstractmethod
    def remove_actor(self, *, actor_name: str):
        """
        Remove an actor
        @params actor_name: actor_name
        """

    @abstractmethod
    def remove_actor_database(self, *, actor_name: str):
        """
        Remove Actor Database
        @params actor_name: actor name
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import timedelta
from enum import Enum

from fim.slivers.network_service import ServiceType


class ErrorCodes(Enum):
    """
    Error code enumeration
    """
    ErrorInvalidArguments = 1
    ErrorInvalidActor = 2
    ErrorInvalidReservation = 3
    ErrorDatabaseError = 4
    ErrorInternalError = 5
    ErrorNoSuchSlice = 6
    ErrorNoSuchResourcePool = 7
    ErrorNoSuchReservation = 8
    ErrorNoSuchBroker = 9
    ErrorInvalidSlice = 10
    ErrorNoSuchActor = 11
    ErrorTransportFailure = 12
    ErrorTransportTimeout = 13
    ErrorNoSuchDelegation = 14
    ErrorInvalidToken = 15
    ErrorSliceExists = 16
    ErrorNoSuchPoa = 17

    def interpret(self, exception=None):
        interpretations = {
            1: "Invalid Arguments",
            2: "Invalid Actor",
            3: "Invalid Reservation",
            4: "Database Error occurred",
            5: "Internal Error occurred",
            6: "No such slice found",
            7: "No such resource pool found",
            8: "No such reservation found",
            9: "No such broker",
            10: "Invalid Slice",
            11: "No such actor",
            12: "Transport failure",
            13: "Transport timeout",
            14: "No such delegation found",
            15: "Invalid Token",
            16: "Slice exists",
            17: "No such poa found"
          }
        if exception is None:
            return interpretations[self.value]
        else:
            return str(exception) + ". " + interpretations[self.value]


class Constants:
    """
    Constants
    """
    MANAGEMENT_API_TIMEOUT_IN_SECONDS = 120.0

    CONTAINER_MANAGMENT_OBJECT_ID = "manager"

    PROPERTY_PICKLE_PROPERTIES = "properties"

    PROPERTY_CONF_KAFKA_POLL_TIMEOUT = "consumer.poll.timeout"
    PROPERTY_CONF_KAFKA_BATCH_SIZE = "commit.batch.size"
    PROPERTY_CONF_KAFKA_AUTO_COMMIT_INTERVAL = "auto.commit.interval.ms"
    PROPERTY_CONF_KAFKA_ENABLE_AUTO_COMMIT = "enable.auto.commit"

    CONFIG_SECTION_RUNTIME = "runtime"
    PROPERTY_CONF_KAFKA_SERVER = "kafka-server"
    PROPERTY_CONF_KAFKA_SCHEMA_REGISTRY = "kafka-schema-registry-url"
    PROPERTY_CONF_KAFKA_KEY_SCHEMA = "kafka-key-schema"
    PROPERTY_CONF_KAFKA_VALUE_SCHEMA = "kafka-value-schema"
    PROPERTY_CONF_KAFKA_S_SL_CA_LOCATION = "kafka-ssl-ca-location"
    PROPERTY_CONF_KAFKA_SSL_CERTIFICATE_LOCATION = "kafka-ssl-certificate-location"
    PROPERTY_CONF_KAFKA_SSL_KEY_LOCATION = "kafka-ssl-key-location"
    PROPERTY_CONF_KAFKA_SSL_KEY_PASSWORD = "kafka-ssl-key-password"
    PROPERTY_CONF_KAFKA_SECURITY_PROTOCOL = "kafka-security-protocol"
    PROPERTY_CONF_KAFKA_GROUP_ID = "kafka-group-id"
    PROPERTY_CONF_KAFKA_SASL_PRODUCER_USERNAME = "kafka-sasl-producer-username"
    PROPERTY_CONF_KAFKA_SASL_PRODUCER_PASSWORD = "kafka-sasl-producer-password"
    PROPERTY_CONF_KAFKA_SASL_CONSUMER_USERNAME = "kafka-sasl-consumer-username"
    PROPERTY_CONF_KAFKA_SASL_CONSUMER_PASSWORD = "kafka-sasl-consumer-password"
    PROPERTY_CONF_KAFKA_SASL_MECHANISM = "kafka-sasl-mechanism"
    PROPERTY_CONF_KAFKA_REQUEST_TIMEOUT_MS = "request.timeout.ms"
    PROPERTY_CONF_KAFKA_MAX_MESSAGE_SIZE = "message.max.bytes"
    PROPERTY_CONF_KAFKA_FETCH_MAX_MESSAGE_SIZE = "fetch.message.max.bytes"

    KAFKA_TOPIC = "kafka-topic"
    NAME = "name"
    TYPE = "type"
    GUID = "guid"
    DESCRIPTION = "description"
    LABEL = "label"
    HANDLER = "handler"
    CREDMGR_HOST = "credmgr-host"
    PUBLISH_INTERVAL = "publish-interval"
    REFRESH_INTERVAL = "refresh-interval"
    DELEGATION = "delegation"

    PROPERTY_CLASS_NAME = "ObjectClassName"
    PROPERTY_MODULE_NAME = "ModuleName"
    PROPERTY_ID = "MOID"
    PROPERTY_TYPE_ID = "MOTYPEID"
    PROPERTY_ACTOR_NAME = "MOActorName"
    PROPERTY_PROXIES_LENGTH = "MOProxiesLength"
    PROPERTY_PROXIES_PREFIX = "MOProxiesPrefix."
    PROPERTY_PROXIES_PROTOCOL = ".protocol"
    PROPERTY_PROXIES_CLASS = ".class"
    PROPERTY_PROXIES_MODULE = ".module"

    PROPERTY_CONF_PROMETHEUS_REST_PORT = "prometheus.port"
    PROPERTY_CONF_CONTROLLER_REST_PORT = "orchestrator.rest.port"
    PROPERTY_CONF_CONTROLLER_CREATE_WAIT_TIME = "orchestrator.create.wait.time"
    PROPERTY_CONF_RPC_REQUEST_TIMEOUT_SECONDS = "rpc.request.timeout.seconds"
    PROPERTY_CONF_RPC_RETRIES = "rpc.retries"

    PROPERTY_SUBSTRATE_FILE = "substrate.file"
    PROPERTY_AGGREGATE_RESOURCE_MODEL = "AggregateResourceModel"

    PROPERTY_CONF_PLUGIN_DIR = "plugin-dir"

    CONFIG_SECTION_PDP = 'pdp'

    CONFIG_LOGGING_SECTION = 'logging'
    PROPERTY_CONF_LOG_FILE = 'log-file'
    PROPERTY_CONF_METRICS_LOG_FILE = 'metrics-log-file'
    PROPERTY_CONF_HANDLER_LOG_FILE = 'handler-log-file'
    PROPERTY_CONF_LOG_LEVEL = 'log-level'
    PROPERTY_CONF_LOG_RETAIN = 'log-retain'
    PROPERTY_CONF_LOG_SIZE = 'log-size'
    PROPERTY_CONF_LOG_DIRECTORY = 'log-directory'
    PROPERTY_CONF_LOGGER = "logger"

    CONFIG_SECTION_CONTAINER = "container"
    PROPERTY_CONF_CONTAINER_GUID = "container.guid"

    CONFIG_SECTION_TIME = "time"
    PROPERTY_CONF_TIME_START_TIME = "time.startTime"
    PROPERTY_CONF_TIME_CYCLE_MILLIS = "time.cycleMillis"
    PROPERTY_CONF_TIME_MANUAL = "time.manual"

    CONFIG_SECTION_O_AUTH = "oauth"
    PROPERTY_CONF_O_AUTH_JWKS_URL = "jwks-url"
    PROPERTY_CONF_O_AUTH_KEY_REFRESH = "key-refresh"
    PROPERTY_CONF_O_AUTH_TRL_REFRESH = "trl-refresh"
    PROPERTY_CONF_O_AUTH_VERIFY_EXP = "verify-exp"

    CONFIG_SECTION_DATABASE = "database"
    PROPERTY_CONF_DB_USER = "db-user"
    PROPERTY_CONF_DB_PASSWORD = "db-password"
    PROPERTY_CONF_DB_NAME = "db-name"
    PROPERTY_CONF_DB_HOST = "db-host"

    CONFIG_SECTION_NEO4J = "neo4j"
    CONFIG_SECTION_BQM = "bqm"

    PROPERTY_CONF_MODULE_NAME = 'module'
    PROPERTY_CONF_CLASS_NAME = 'class'
    PROPERTY_CONF_PROPERTIES_NAME = 'properties'

    CONFIG_SECTION_ACTOR = "actor"

    PROTOCOL_LOCAL = "local"
    PROTOCOL_KAFKA = "kafka"

    HOME_DIRECTORY = '/usr/src/app/'
    SUPERBLOCK_LOCATION = HOME_DIRECTORY + "state_recovery.lock"
    MAINTENANCE_LOCATION = HOME_DIRECTORY + "maintenance.lock"
    MODEL_RELOAD_LOCATION = HOME_DIRECTORY + "reload.model"
    CONFIGURATION_FILE = "/etc/fabric/actor/config/config.yaml"
    STATE_FILE_LOCATION = '/tmp/fabric_actor.tmp'
    MAINT_PROJECT_ID = 'maint.project.id'
    INFRASTRUCTURE_PROJECT_ID = "infrastructure.project.id"

    ELASTIC_TIME = "request.elasticTime"
    ELASTIC_SIZE = "request.elasticSize"

    QUERY_ACTION_DISCOVER_BQM = "discover.bqm"
    QUERY_ACTION = "query.action"
    QUERY_RESPONSE = "query.response"
    QUERY_RESPONSE_STATUS = "query.response.status"
    QUERY_RESPONSE_MESSAGE = "query.response.message"
    QUERY_DETAIL_LEVEL = "query.detail.level"
    BROKER_QUERY_MODEL = "bqm"
    BROKER_QUERY_MODEL_FORMAT = "bqm.format"
    POOL_TYPE = "neo4j"

    UNIT_MODIFY_PROP_MESSAGE_SUFFIX = ".message"
    UNIT_MODIFY_PROP_CODE_SUFFIX = ".code"
    UNIT_MODIFY_PROP_PREFIX = "unit.modify."

    SASL_USERNAME = 'sasl.username'
    SASL_PASSWORD = 'sasl.password'
    SASL_MECHANISM = 'sasl.mechanism'
    BOOTSTRAP_SERVERS = 'bootstrap.servers'
    SECURITY_PROTOCOL = 'security.protocol'
    GROUP_ID = 'group.id'
    SSL_CA_LOCATION = 'ssl.ca.location'
    SSL_CERTIFICATE_LOCATION = 'ssl.certificate.location'
    SSL_KEY_LOCATION = 'ssl.key.location'
    SSL_KEY_PASSWORD = 'ssl.key.password'
    SCHEMA_REGISTRY_URL = 'schema.registry.url'

    INVALID_ARGUMENT = "Invalid argument"
    UNINITIALIZED_STATE = "Uninitialized state"
    INVALID_ACTOR_STATE = "Invalid state, actor cannot receive calls"
    OBJECT_NOT_FOUND = "{} not found in datatabase e: {}"
    FAILED_TO_ADD = "Failed to add {} to database e: {}"
    FAILED_TO_REMOVE = "Failed to remove {} from database e: {}"
    NOT_SPECIFIED_PREFIX = "Not {} specified"
    RECOVERY = "[recovery]"
    ISSUE_OPERATION = "Issued {} request for reservation #{} State={}"
    INVALID_PENDING_STATE = "Invalid pending state"
    RESTARTING_ACTIONS = "Restarting configuration actions for reservation #{}"
    RESTARTING_ACTIONS_COMPLETE = "Restarting configuration actions for reservation #{} complete"
    EXTEND_TICKET = "[extend ticket]"

    MANAGEMENT_INTER_ACTOR_OUTBOUND_MESSAGE = "Outbound [Kafka] [Management] Message {} written to {}"
    MANAGEMENT_INTER_ACTOR_INBOUND_MESSAGE = "Inbound [Kafka] [Management] Message received {}"
    MANAGEMENT_API_TIMEOUT_OCCURRED = "Management API Timeout Occurred"
    MANAGEMENT_INTER_ACTOR_MESSAGE_FAILED = "Outbound [Kafka] [Management] Message {} could not be written to {}"

    INVALID_MANAGEMENT_OBJECT_TYPE = "Invalid Management Object type {}"
    NO_POOL = "there is no pool to satisfy this request"
    INVALID_RECOVERY_STATE = "This state should not be reached during recovery"

    INVALID_STATE = "Invalid state"
    INVALID_IP = "Invalid ip address: {}"
    EXCEPTION_OCCURRED = "Exception occurred {}"
    NOT_IMPLEMENTED = "Not Implemented"

    UNSUPPORTED_RESOURCE_TYPE = "Unsupported resource type: {}"
    CLOSURE_BY_TICKET_REVIEW_POLICY = "TicketReviewPolicy: Closing reservation due to failure in slice"
    MAINTENANCE_MODE_ERROR = "Testbed is in maintenance mode: Create, Modify and Renew Slice(s) are disabled!"
    PENDING_OPERATION_ERROR = "Reservation has a pending operation"

    CLAIMS_SUB = "sub"
    CLAIMS_EMAIL = "email"
    CLAIMS_PROJECTS = "projects"
    UUID = "uuid"
    TAGS = "tags"
    TOKEN_HASH = "token_hash"
    PROJECT_ID = "project_id"
    USERS = "users"
    DEADLINE = "deadline"

    PROPERTY_EXCEPTION_MESSAGE = "exception.message"
    PROPERTY_TARGET_NAME = "target.name"
    PROPERTY_TARGET_RESULT_CODE = "target.code"
    PROPERTY_TARGET_RESULT_CODE_MESSAGE = "target.code.message"
    PROPERTY_ACTION_SEQUENCE_NUMBER = "action.sequence"
    PROPERTY_POA_INFO = "poa_info"
    PROPERTY_INFO = "info"
    PROPERTY_CODE = "code"
    PROPERTY_MESSAGE = "message"
    POA_ID = "poa_id"

    RESULT_CODE_EXCEPTION = -1
    RESULT_CODE_OK = 0
    TARGET_CREATE = "create"
    TARGET_DELETE = "delete"
    TARGET_MODIFY = "modify"
    TARGET_POA = "poa"
    TARGET_CLEAN_RESTART = "clean_restart"

    RSV_SLC_ID = 'rsv_slc_id'
    DLG_SLC_ID = 'dlg_slc_id'

    USER_SSH_KEY = "user.ssh.key"
    ALGORITHM = 'algorithm'

    # Orchestrator Lease params
    TWO_WEEKS = timedelta(days=15)
    DEFAULT_MAX_DURATION = TWO_WEEKS
    LEASE_TIME_FORMAT = "%Y-%m-%d %H:%M:%S %z"
    DEFAULT_LEASE_IN_HOURS = 24
    LONG_LIVED_SLICE_TIME_WEEKS = timedelta(weeks=26)

    ALL = "ALL"
    AL2S = "AL2S"
    PEERED = "Peered"

    # Default offset used to pick a VLAN from the range 1-4096
    DEFAULT_VLAN_OFFSET = 10
    VLAN_START = 1
    VLAN_END = 4096

    CONFIG_PROPERTIES_FILE = "config.properties.file"
    MODE = "mode"

    OPENSTACK_VNIC_MODEL = "OpenStack-vNIC"

    INTERNAL_SERVER_ERROR_MAINT_MODE = 501

    L3_SERVICES = [ServiceType.FABNetv6Ext, ServiceType.FABNetv4Ext, ServiceType.FABNetv4,
                   ServiceType.FABNetv6, ServiceType.L3VPN]

    L3_FABNET_SERVICES = [ServiceType.FABNetv6Ext, ServiceType.FABNetv4Ext, ServiceType.FABNetv4,
                          ServiceType.FABNetv6]

    L3_FABNET_SERVICES_STR = [str(ServiceType.FABNetv6), str(ServiceType.FABNetv4), str(ServiceType.FABNetv4Ext),
                              str(ServiceType.FABNetv6Ext)]

    L3_FABNET_EXT_SERVICES = [ServiceType.FABNetv6Ext, ServiceType.FABNetv4Ext]

    L3_FABNET_NON_EXT_SERVICES = [ServiceType.FABNetv4, ServiceType.FABNetv6]

    L3_FABNETv6_SERVICES = [ServiceType.FABNetv6Ext, ServiceType.FABNetv6]

    L3_FABNETv4_SERVICES = [ServiceType.FABNetv4Ext, ServiceType.FABNetv4]

    SUPPORTED_SERVICES_STR = [str(ServiceType.L2STS), str(ServiceType.L2Bridge), str(ServiceType.L2PTP),
                              str(ServiceType.FABNetv4), str(ServiceType.FABNetv6), str(ServiceType.PortMirror),
                              str(ServiceType.FABNetv4Ext), str(ServiceType.FABNetv6Ext), str(ServiceType.L3VPN)]

    IGNORABLE_NS = [ServiceType.P4, ServiceType.OVS, ServiceType.MPLS, ServiceType.VLAN]

    SUPPORTED_SERVICES = [ServiceType.L2STS, ServiceType.L2Bridge, ServiceType.L2PTP, ServiceType.FABNetv6,
                          ServiceType.FABNetv4, ServiceType.PortMirror, ServiceType.FABNetv4Ext,
                          ServiceType.FABNetv6Ext, ServiceType.L3VPN]

    SLICE_NO_LIMIT_LIFETIME = "Slice.NoLimitLifetime"



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from datetime import datetime
from typing import TYPE_CHECKING, List

from fabric_mb.message_bus.messages.delegation_avro import DelegationAvro
from fabric_mb.message_bus.messages.broker_query_model_avro import BrokerQueryModelAvro
from fabric_mb.message_bus.messages.poa_avro import PoaAvro
from fabric_mb.message_bus.messages.reservation_predecessor_avro import ReservationPredecessorAvro
from fabric_mb.message_bus.messages.ticket_reservation_avro import TicketReservationAvro
from fabric_mb.message_bus.messages.unit_avro import UnitAvro
from fim.slivers.base_sliver import BaseSliver
from fim.user import GraphFormat

from fabric_cf.actor.core.common.exceptions import ManageException
from fabric_cf.actor.core.manage.controller_management_object import ControllerManagementObject
from fabric_cf.actor.core.apis.abc_mgmt_controller_mixin import ABCMgmtControllerMixin
from fabric_cf.actor.core.manage.local.local_actor import LocalActor
from fabric_cf.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.proxy_avro import ProxyAvro
    from fabric_mb.message_bus.messages.reservation_mng import ReservationMng

    from fabric_cf.actor.core.apis.abc_client_actor_management_object import ABCClientActorManagementObject
    from fabric_cf.actor.security.auth_token import AuthToken


class LocalController(LocalActor, ABCMgmtControllerMixin):
    def __init__(self, *, manager: ABCClientActorManagementObject, auth: AuthToken):
        super().__init__(manager=manager, auth=auth)

        if not isinstance(manager, ControllerManagementObject):
            raise ManageException("Invalid manager object. Required: {}".format(type(ControllerManagementObject)))

    def add_broker(self, *, broker: ProxyAvro) -> bool:
        self.clear_last()
        try:
            result = self.manager.add_broker(broker=broker, caller=self.auth)
            self.last_status = result

            return result.get_code() == 0
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return False

    def update_broker(self, *, broker: ProxyAvro) -> bool:
        self.clear_last()
        try:
            result = self.manager.update_broker(broker=broker, caller=self.auth)
            self.last_status = result

            return result.get_code() == 0
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return False

    def get_brokers(self, *, broker: ID = None, id_token: str = None) -> List[ProxyAvro]:
        self.clear_last()
        try:
            result = self.manager.get_brokers(caller=self.auth, broker_id=broker)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.proxies
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

    def get_broker_query_model(self, *, broker: ID, id_token: str, level: int,
                               graph_format: GraphFormat) -> BrokerQueryModelAvro:
        self.clear_last()
        try:
            result = self.manager.get_broker_query_model(broker=broker, caller=self.auth, id_token=id_token,
                                                         level=level, graph_format=graph_format)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return result.model
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

    def claim_delegations(self, *, broker: ID, did: ID, id_token: str = None) -> DelegationAvro:
        self.clear_last()
        try:
            result = self.manager.claim_delegations(broker=broker, did=did, caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return self.get_first(result_list=result.delegations)
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

    def reclaim_delegations(self, *, broker: ID, did: ID, id_token: str = None) -> DelegationAvro:
        self.clear_last()
        try:
            result = self.manager.reclaim_delegations(broker=broker, did=did, caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return self.get_first(result_list=result.delegations)
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

    def get_reservation_units(self, *, rid: ID, id_token: str = None) -> List[UnitAvro]:
        self.clear_last()
        try:
            result = self.manager.get_reservation_units(caller=self.auth, rid=rid)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.units
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

    def add_reservation(self, *, reservation: TicketReservationAvro) -> ID:
        self.clear_last()
        try:
            result = self.manager.add_reservation(reservation=reservation, caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return ID(uid=result.get_result())
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

    def add_reservations(self, *, reservations: List[TicketReservationAvro]) -> List[ID]:
        self.clear_last()
        try:
            result = self.manager.add_reservations(reservations=reservations, caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                rids = []
                for r in result.result:
                    rids.append(ID(uid=r))

                return rids
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

    def demand_reservation_rid(self, *, rid: ID) -> bool:
        self.clear_last()
        try:
            result = self.manager.demand_reservation_rid(rid=rid, caller=self.auth)
            self.last_status = result

            return result.get_code() == 0
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return False

    def demand_reservation(self, *, reservation: ReservationMng) -> bool:
        self.clear_last()
        try:
            result = self.manager.demand_reservation(reservation=reservation, caller=self.auth)
            self.last_status = result

            return result.get_code() == 0
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return False

    def extend_reservation(self, *, reservation: ID, new_end_time: datetime, sliver: BaseSliver,
                           dependencies: List[ReservationPredecessorAvro] = None) -> bool:
        self.clear_last()
        try:
            result = self.manager.extend_reservation(reservation=reservation,
                                                     new_end_time=new_end_time, sliver=sliver,
                                                     caller=self.auth, dependencies=dependencies)
            self.last_status = result

            return result.get_code() == 0
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return False

    def clone(self):
        return LocalController(manager=self.manager, auth=self.auth)

    def modify_reservation(self, *, rid: ID, modified_sliver: BaseSliver) -> bool:
        self.clear_last()
        try:
            result = self.manager.modify_reservation(rid=rid, modified_sliver=modified_sliver, caller=self.auth)
            self.last_status = result

            return result.get_code() == 0
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return False

    def poa(self, *, poa: PoaAvro) -> bool:
        self.clear_last()
        try:
            result = self.manager.poa(poa=poa, caller=self.auth)
            self.last_status = result

            return result.get_code() == 0
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return False


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric_cf.actor.core.apis.abc_authority_proxy import ABCAuthorityProxy
from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
from fabric_cf.actor.core.apis.abc_client_callback_proxy import ABCClientCallbackProxy
from fabric_cf.actor.core.apis.abc_controller_callback_proxy import ABCControllerCallbackProxy
from fabric_cf.actor.core.apis.abc_controller_reservation import ABCControllerReservation
from fabric_cf.actor.core.apis.abc_rpc_request_state import ABCRPCRequestState
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.security.auth_token import AuthToken
from fabric_cf.actor.test.dummy_proxy import DummyProxy


class DummyAuthorityProxy(DummyProxy, ABCAuthorityProxy):
    def __init__(self, *, auth: AuthToken = None):
        super().__init__(auth)

    def prepare_redeem(self, *, reservation: ABCControllerReservation, callback: ABCControllerCallbackProxy, caller:
    AuthToken) -> ABCRPCRequestState:
        return None

    def prepare_extend_lease(self, *, reservation: ABCControllerReservation, callback: ABCControllerCallbackProxy,
                             caller: AuthToken) -> ABCRPCRequestState:
        return None

    def prepare_modify_lease(self, *, reservation: ABCControllerReservation, callback: ABCControllerCallbackProxy,
                             caller: AuthToken) -> ABCRPCRequestState:
        return None

    def prepare_close(self, *, reservation: ABCControllerReservation, callback: ABCControllerCallbackProxy,
                      caller: AuthToken) -> ABCRPCRequestState:
        return None

    def prepare_ticket(self, *, reservation: ABCReservationMixin, callback: ABCClientCallbackProxy, caller: AuthToken) -> ABCRPCRequestState:
        return None

    def prepare_claim(self, *, reservation: ABCReservationMixin, callback: ABCClientCallbackProxy, caller: AuthToken) -> ABCRPCRequestState:
        return None

    def prepare_reclaim(self, *, reservation: ABCReservationMixin, callback: ABCClientCallbackProxy, caller: AuthToken) -> ABCRPCRequestState:
        return None

    def prepare_extend_ticket(self, *, reservation: ABCReservationMixin, callback: ABCClientCallbackProxy, caller: AuthToken) -> ABCRPCRequestState:
        return None

    def prepare_relinquish(self, *, reservation: ABCReservationMixin, callback: ABCClientCallbackProxy, caller: AuthToken) -> ABCRPCRequestState:
        return None

    def prepare_query(self, *, callback: ABCCallbackProxy, query:dict, caller: AuthToken):
        return None


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, List, Tuple, Dict

from fabric_mb.message_bus.messages.delegation_avro import DelegationAvro
from fabric_mb.message_bus.messages.poa_avro import PoaAvro
from fabric_mb.message_bus.messages.poa_info_avro import PoaInfoAvro
from fabric_mb.message_bus.messages.site_avro import SiteAvro

from fabric_cf.actor.core.apis.abc_component import ABCComponent
from fabric_cf.actor.core.container.maintenance import Site

if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
    from fabric_mb.message_bus.messages.reservation_state_avro import ReservationStateAvro
    from fabric_mb.message_bus.messages.slice_avro import SliceAvro
    from fabric_cf.actor.core.util.id import ID


class ABCMgmtActor(ABCComponent):
    @abstractmethod
    def get_slices(self, *, slice_id: ID = None, slice_name: str = None, email: str = None, project: str = None,
                   states: List[int] = None, limit: int = None, offset: int = None,
                   user_id: str = None) -> List[SliceAvro] or None:
        """
        Obtains all slices.
        @param slice_id slice id
        @param slice_name slice name
        @param email email
        @param project project id
        @param states slice states
        @param limit limit
        @param offset offset
        @param user_id user_id
        @return returns list of slices
        """

    @abstractmethod
    def add_slice(self, *, slice_obj: SliceAvro) -> ID:
        """
        Adds a new slice
        @param slice_obj slice
        @return returns slice id
        """

    @abstractmethod
    def remove_slice(self, *, slice_id: ID) -> bool:
        """
        Removes the specified slice
        @param slice_id slice id
        @return true for success; false otherwise
        """

    @abstractmethod
    def delete_slice(self, *, slice_id: ID) -> bool:
        """
        Deletes the specified slice - moves the slice into Closing state.
        @param slice_id slice ID
        @return true for success; false otherwise
        """

    @abstractmethod
    def update_slice(self, *, slice_obj: SliceAvro, modify_state: bool = False) -> bool:
        """
        Updates the specified slice.
        The only updatable slice attributes are:
        - description
        - all properties lists
        @param slice_obj slice
        @param modify_state - trigger state change
        @return true for success; false otherwise
        """

    @abstractmethod
    def accept_update_slice(self, *, slice_id: ID) -> bool:
        """
        Accept the last slice update
        @param slice_id slice_id
        @return true for success; false otherwise
        """

    @abstractmethod
    def get_reservations(self, *, states: List[int] = None, slice_id: ID = None,
                         rid: ID = None, oidc_claim_sub: str = None, email: str = None, rid_list: List[str] = None,
                         type: str = None, site: str = None, node_id: str = None) -> List[ReservationMng]:
        """
        Get Reservations
        @param states states
        @param slice_id slice ID
        @param rid reservation id
        @param oidc_claim_sub: oidc claim sub
        @param email: user email
        @param rid_list: list of Reservation Id
        @param type type of reservations like NodeSliver/NetworkServiceSliver
        @param site site
        @param node_id node id
        Obtains all reservations
        @return returns list of the reservations
        """

    @abstractmethod
    def get_sites(self, *, site: str) -> List[SiteAvro] or None:
        """
        @param site site name
        Obtains Maintenance Info for a site
        @return returns list of the Sites
        """

    @abstractmethod
    def update_reservation(self, *, reservation: ReservationMng) -> bool:
        """
        Updates the specified reservation
        The fields that can be updated are:

        - all properties lists

        @param reservation reservation to be updated
        @return true for success; false otherwise
        """

    @abstractmethod
    def close_reservation(self, *, rid: ID) -> bool:
        """
        Closes the specified reservation
        @param rid reservation id
        @return true for success; false otherwise
        """

    @abstractmethod
    def close_delegation(self, *, did: str) -> bool:
        """
        Closes the specified delegation
        @param did delegation id
        @return true for success; false otherwise
        """

    @abstractmethod
    def close_reservations(self, *, slice_id: ID) -> bool:
        """
        Closes all reservations in the specified slice.
        @param slice_id slice ID
        @return true for success; false otherwise
        """

    @abstractmethod
    def remove_reservation(self, *, rid: ID) -> bool:
        """
        Removes the specified reservation.
        Note only closed reservations can be removed.
        @param rid reservation id of the reservation to be removed
        @return true for success; false otherwise
        """

    @abstractmethod
    def remove_delegation(self, *, did: str) -> bool:
        """
        Removes the specified delegation.
        Note only closed delegation can be removed.
        @param did delegation id of the delegation to be removed
        @return true for success; false otherwise
        """

    @abstractmethod
    def get_reservation_state_for_reservations(self, *, reservation_list: List[str]) -> List[ReservationStateAvro]:
        """
        Returns the state of each of the specified reservations.
        The order in the return list matches the order in the @reservations li
        @param reservation_list list of reservations
        @return list of state of the specified reservations
        """

    @abstractmethod
    def get_name(self) -> str:
        """
        Returns the name of the actor.
        @return returns name of the actor
        """

    @abstractmethod
    def get_guid(self) -> ID:
        """
        Returns the guid of the actor.
        @return returns guid of the actor
        """

    @abstractmethod
    def clone(self):
        """
        Creates clone of this proxy to make it possible to be used by a separate thread.
        @return returns the cloned actor
        """

    @abstractmethod
    def get_delegations(self, *, slice_id: ID = None, states: List[int] = None,
                        delegation_id: str = None) -> List[DelegationAvro]:
        """
        Get Delegations
        @param slice_id slice id
        @param states states
        @param delegation_id delegation id
        @return returns list of the delegations
        """

    def is_testbed_in_maintenance(self) -> Tuple[bool, Dict[str, str] or None]:
        """
        Determine if testbed is in maintenance
        @return True if testbed is in maintenance; False otherwise
        """
        return False, None

    def is_site_in_maintenance(self, *, site_name: str) -> Tuple[bool, Site or None]:
        """
        Determine if site is in maintenance
        @return True if site is in maintenance; False otherwise
        """
        return False, None

    def is_sliver_provisioning_allowed(self, *, project: str, email: str, site: str,
                                       worker: str) -> Tuple[bool, str or None]:
        """
        Determine if sliver can be provisioned
        Sliver provisioning can be prohibited if Testbed or Site or Worker is in maintenance mode
        Sliver provisioning in maintenance mode may be allowed for specific projects/users
        @param project project
        @param email user's email
        @param site site name
        @param worker worker name
        @return True if allowed; False otherwise
        """
        return True, None

    def is_slice_provisioning_allowed(self, *, project: str, email: str) -> bool:
        """
        Determine if slice can be provisioned
        Slice provisioning can be prohibited if Testbed is in maintenance mode
        Slice provisioning in maintenance mode may be allowed for specific projects/users
        @param project project
        @param email user's email
        @return True if allowed; False otherwise
        """
        return True

    @abstractmethod
    def get_poas(self, *, states: List[int] = None, slice_id: ID = None, rid: ID = None,
                 email: str = None, poa_id: str = None, project_id: str = None,
                 limit: int = 200, offset: int = 0) -> List[PoaInfoAvro]:
        """
        Get POA
        @param states states
        @param slice_id slice ID
        @param rid reservation id
        @param email: user email
        @param project_id: project_id
        @param poa_id: poa_id
        @param limit: limit of records to be returned
        @param offset: offset
        @return returns list of the poas
        """


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from enum import Enum


class ReservationStates(Enum):
    """
    Reservation states
    """
    Nascent = 1
    Ticketed = 2
    Active = 4
    ActiveTicketed = 5
    Closed = 6
    CloseWait = 7
    Failed = 8
    Unknown = 9
    CloseFail = 10

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    @staticmethod
    def translate(state_name: str):
        if state_name.lower() == ReservationStates.Nascent.name.lower():
            return ReservationStates.Nascent
        elif state_name.lower() == ReservationStates.Ticketed.name.lower():
            return ReservationStates.Ticketed
        elif state_name.lower() == ReservationStates.Active.name.lower():
            return ReservationStates.Active
        elif state_name.lower() == ReservationStates.ActiveTicketed.name.lower():
            return ReservationStates.ActiveTicketed
        elif state_name.lower() == ReservationStates.Closed.name.lower():
            return ReservationStates.Closed
        elif state_name.lower() == ReservationStates.CloseWait.name.lower():
            return ReservationStates.CloseWait
        elif state_name.lower() == ReservationStates.Failed.name.lower():
            return ReservationStates.Failed
        elif state_name.lower() == ReservationStates.CloseFail.name.lower():
            return ReservationStates.Failed
        else:
            return ReservationStates.Unknown


class ReservationPendingStates(Enum):
    """
    Pending operation states
    """
    None_ = 11
    Ticketing = 12
    Redeeming = 14
    ExtendingTicket = 15
    ExtendingLease = 16
    Priming = 17
    Blocked = 18
    Closing = 19
    Probing = 20
    ClosingJoining = 21
    ModifyingLease = 22
    AbsorbUpdate = 23
    SendUpdate = 24
    PrimingPoa = 25
    WaitingPoaResponse = 26
    Unknown = 27

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class JoinState(Enum):
    """
    Join states
    """
    None_ = 31
    NoJoin = 32
    BlockedJoin = 33
    BlockedRedeem = 34
    Joining = 35
    BlockedTicket = 36
    BlockedExtendTicket = 37
    BlockedExtendLease = 38

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import time

from fabric_cf.actor.core.apis.abc_controller_policy import ABCControllerPolicy
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.kernel.reservation_client import ClientReservationFactory
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.kernel.slice import SliceFactory
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.resource_type import ResourceType
from fabric_cf.actor.test.core.policy.controller_simple_policy_test import ControllerSimplePolicyTest
from fabric_cf.actor.test.core.policy.controller_ticket_review_policy_test_wrapper import \
    ControllerTicketReviewPolicyTestWrapper


class ControllerTicketReviewPolicyTest(ControllerSimplePolicyTest):
    def get_controller_policy(self) -> ABCControllerPolicy:
        return ControllerTicketReviewPolicyTestWrapper()

    def _test_c_fail(self):
        controller = self.get_controller()
        clock = controller.get_actor_clock()
        Term.clock = clock

        resources = ResourceSet(units=1, rtype=ResourceType(resource_type="1"))
        slice_obj = SliceFactory.create(slice_id=ID(), name="fail")
        controller.register_slice(slice_object=slice_obj)

        start = 5
        end = 10

        term = Term(start=clock.cycle_start_date(cycle=start), end=clock.cycle_end_date(cycle=end))

        r1 = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        r1.set_renewable(renewable=False)
        controller.register(reservation=r1)
        controller.demand(rid=r1.get_reservation_id())

        r2 = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        r2.set_renewable(renewable=False)
        controller.register(reservation=r2)
        controller.demand(rid=r2.get_reservation_id())

        for i in range(1, end + 3):
            controller.external_tick(cycle=i)

            while controller.get_current_cycle() != i:
                time.sleep(0.001)

            if i >= start and (i < (end - 1)):
                self.assertTrue(r1.is_closed())
                self.assertTrue(r2.is_closed())
                self.assertTrue(r2.get_notices().__contains__(Constants.CLOSURE_BY_TICKET_REVIEW_POLICY))

    def _test_d_nascent(self):
        controller = self.get_controller()
        clock = controller.get_actor_clock()
        Term.clock = clock

        resources = ResourceSet(units=1, rtype=ResourceType(resource_type="1"))
        slice_obj = SliceFactory.create(slice_id=ID(), name="nascent")
        controller.register_slice(slice_object=slice_obj)

        start = 5
        end = 10

        term = Term(start=clock.cycle_start_date(cycle=start), end=clock.cycle_end_date(cycle=end))

        r1 = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        r1.set_renewable(renewable=False)
        controller.register(reservation=r1)
        controller.demand(rid=r1.get_reservation_id())

        r2 = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        r2.set_renewable(renewable=False)
        controller.register(reservation=r2)

        r2demanded = False

        for i in range(1, end + 3):
            controller.external_tick(cycle=i)

            while controller.get_current_cycle() != i:
                time.sleep(0.001)

            if i == (start -3) and not r2demanded:
                self.assertTrue(r1.is_ticketed())
                self.assertTrue(r2.is_nascent())
                controller.demand(rid=r2.get_reservation_id())
                r2demanded = True

            if i >= start and (i < end - 1):
                self.assertTrue(r1.is_active())
                self.assertTrue(r2.is_active())

            if i > end:
                self.assertTrue(r1.is_closed())
                self.assertTrue(r2.is_closed())

    def _test_e_fail_and_nascent(self):
        controller = self.get_controller()
        clock = controller.get_actor_clock()
        Term.clock = clock

        resources = ResourceSet(units=1, rtype=ResourceType(resource_type="1"))
        slice_obj = SliceFactory.create(slice_id=ID(), name="fail_nascent")
        controller.register_slice(slice_object=slice_obj)

        start = 10
        end = 15

        term = Term(start=clock.cycle_start_date(cycle=start), end=clock.cycle_end_date(cycle=end))

        r1 = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        r1.set_renewable(renewable=False)
        controller.register(reservation=r1)
        controller.demand(rid=r1.get_reservation_id())

        r2 = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        r2.set_renewable(renewable=False)
        controller.register(reservation=r2)
        r2demanded = False

        r3 = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        r3.set_renewable(renewable=False)
        controller.register(reservation=r3)
        controller.demand(rid=r3.get_reservation_id())

        for i in range(1, end + 3):
            controller.external_tick(cycle=i)

            while controller.get_current_cycle() != i:
                time.sleep(0.001)

            if i > 2 and not r2demanded:
                self.assertTrue(r1.is_failed())
                self.assertTrue(r2.is_nascent())
                self.assertTrue(r3.is_ticketed())

                if i > 6:
                    controller.demand(rid=r2.get_reservation_id())
                    r2demanded = True

            if (i >= start) and (i < (end - 1)):
                self.assertTrue(r1.is_closed())
                self.assertTrue(r2.is_closed())
                self.assertTrue(r3.is_closed())
                self.assertTrue(r2.get_notices().__contains__(Constants.CLOSURE_BY_TICKET_REVIEW_POLICY))
                self.assertTrue(r3.get_notices().__contains__(Constants.CLOSURE_BY_TICKET_REVIEW_POLICY))

            if i > end:
                self.assertTrue(r1.is_closed())
                self.assertTrue(r2.is_closed())
                self.assertTrue(r3.is_closed())
                self.assertTrue(r2.get_notices().__contains__(Constants.CLOSURE_BY_TICKET_REVIEW_POLICY))
                self.assertTrue(r3.get_notices().__contains__(Constants.CLOSURE_BY_TICKET_REVIEW_POLICY))



# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import datetime

from fim.slivers.base_sliver import BaseSliver

from fabric_cf.actor.core.apis.abc_base_plugin import ABCBasePlugin
from fabric_cf.actor.core.apis.abc_authority_proxy import ABCAuthorityProxy
from fabric_cf.actor.core.apis.abc_concrete_set import ABCConcreteSet
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import TicketException
from fabric_cf.actor.core.delegation.resource_ticket import ResourceTicket
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.notice import Notice
from fabric_cf.actor.core.util.resource_type import ResourceType


class Ticket(ABCConcreteSet):
    """
    Ticket is an IConcreteSet implementation that wraps a ResourceDelegation for use inside of a ResourceSet
    """

    def __init__(self, *, resource_ticket: ResourceTicket = None, plugin: ABCBasePlugin = None,
                 authority: ABCAuthorityProxy = None, delegation_id: str = None):
        # Persistent fields
        # The encapsulated resource ticket.
        self.resource_ticket = resource_ticket
        # Units we used to have before the current extend
        self.old_units = 0
        # The delegation from which this ticket was issued
        self.delegation_id = delegation_id

        # Non persistent fields
        # The plugin object
        self.plugin = plugin
        self.logger = None
        if plugin is not None:
            self.logger = plugin.get_logger()
        # The authority who owns the resources described in this concrete set
        self.authority = authority
        # The reservation this ticket belongs to
        self.reservation = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['plugin']
        del state['logger']
        del state['reservation']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.plugin = None
        self.logger = None
        self.reservation = None

    def __str__(self):
        result = f"Ticket [delegation_id= {self.delegation_id} units = {self.get_units()} oldUnits = {self.old_units} "
        if self.reservation is not None:
            slice_obj = self.reservation.get_slice()
            if slice is None:
                self.logger.error("reservation inside ticket has no slice")
            else:
                result += f" Slice={slice_obj.get_name()}"
        result += "]"
        return result

    def restore(self, *, plugin: ABCBasePlugin, reservation: ABCReservationMixin):
        """
        Restore members after instantiating the object post database read
        @param plugin plugin
        @param reservation reservation
        """
        self.plugin = plugin
        self.logger = self.plugin.get_logger()
        self.reservation = reservation

    def get_type(self) -> ResourceType:
        """
        Return resource type
        @return resource type
        """
        if self.resource_ticket is None:
            return None
        return self.resource_ticket.get_type()

    def get_ticket(self) -> ResourceTicket:
        """
        Return resource delegation
        @return resource delegation
        """
        return self.resource_ticket

    def add(self, *, concrete_set: ABCConcreteSet, configure: bool):
        raise TicketException("add() is not supported by Ticket")

    def change(self, *, concrete_set: ABCConcreteSet, configure: bool):
        self.old_units = self.get_units()

        if not isinstance(concrete_set, Ticket):
            raise TicketException(Constants.INVALID_ARGUMENT)

        assert concrete_set.resource_ticket is not None

        self.resource_ticket = concrete_set.resource_ticket.clone()

    def _clone(self):
        result = Ticket(resource_ticket=self.resource_ticket, plugin=self.plugin, authority=self.authority,
                        delegation_id=self.delegation_id)
        result.old_units = self.old_units
        return result

    def clone(self):
        return self._clone()

    def clone_empty(self):
        return self._clone()

    def close(self):
        return

    def collect_released(self):
        return

    def get_notices(self) -> Notice:
        return None

    def get_properties(self) -> dict:
        """
        Returns the ticket properties.
        @returns ticket properties
        """
        if self.resource_ticket is None:
            return None
        return self.resource_ticket.get_properties()

    def get_plugin(self) -> ABCBasePlugin:
        """
        Returns Actor Plugin
        @returns actor plugin
        """
        return self.plugin

    def get_site_proxy(self) -> ABCAuthorityProxy:
        """
        Return corresponding Authority
        @return authority
        """
        return self.authority

    def get_term(self) -> Term:
        """
        Get term
        :return: term
        """
        if self.resource_ticket is None:
            return None
        return self.resource_ticket.get_term()

    def holding(self, *, when: datetime) -> int:
        if when is None:
            raise TicketException(Constants.INVALID_ARGUMENT)

        term = self.get_term()
        if term is None:
            return 0

        if when < term.get_new_start_time():
            if when < term.get_start_time():
                # date is before start time
                return 0
            else:
                # date is in [start, newStart)
                return self.old_units
        else:
            if when > term.get_end_time():
                # date is after end time
                return 0
            else:
                # date is in [newStart,end]
                return self.get_units()

    def is_active(self):
        # valid tickets are always active, if anyone asks
        return True

    def modify(self, *, sliver: BaseSliver):
        raise TicketException("Not supported by TicketSet")

    def probe(self):
        return

    def remove(self, *, concrete_set, configure: bool):
        raise TicketException("Not supported by TicketSet")

    def setup(self, *, reservation: ABCReservationMixin):
        """
        Indicates that we're committing resources to a client (on an an agent).
        May need to touch TicketSet database since we're committing it. On a
        client (orchestrator) this indicates that we have successfully scored
        a ticket. The ticket has already been validated with validate().
        @params reservation the slice for the reservation
        """
        self.reservation = reservation

    def validate_concrete(self, *, rtype: ResourceType, units: int, term: Term):
        if self.get_units() < units:
            raise TicketException("Ticket not valid for requested units")

    def validate_incoming(self):
        return

    def validate_outgoing(self):
        return

    def restart_actions(self):
        return

    def get_units(self) -> int:
        if self.resource_ticket is None:
            return 0
        return self.resource_ticket.get_units()

    def get_delegation_id(self) -> str:
        """
        Get Delegation Id
        :return: delegation id
        """
        return self.delegation_id

    def set_delegation_id(self, *, delegation_id: str):
        """
        Set delgation id
        :param delegation_id: delegation id
        :return:
        """
        self.delegation_id = delegation_id



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import List, Tuple, Any

from fabric_mb.message_bus.messages.abc_message_avro import AbcMessageAvro
from fabric_mb.message_bus.messages.request_by_id_record import RequestByIdRecord
from fabric_mb.message_bus.messages.result_avro import ResultAvro
from fabric_mb.message_bus.messages.auth_avro import AuthAvro
from fabric_mb.message_bus.producer import AvroProducerApi

from fabric_cf.actor.core.common.constants import Constants, ErrorCodes
from fabric_cf.actor.core.common.exceptions import ManageException
from fabric_cf.actor.core.manage.error import Error
from fabric_cf.actor.core.apis.abc_component import ABCComponent
from fabric_cf.actor.core.manage.kafka.kafka_mgmt_message_processor import KafkaMgmtMessageProcessor
from fabric_cf.actor.core.manage.messages.protocol_proxy_mng import ProtocolProxyMng
from fabric_cf.actor.core.util.id import ID


class KafkaProxy(ABCComponent):
    def __init__(self, *, guid: ID, kafka_topic: str, auth: AuthAvro, logger,
                 message_processor: KafkaMgmtMessageProcessor, producer: AvroProducerApi = None):
        self.management_id = guid
        self.auth = auth
        self.logger = logger
        self.kafka_topic = kafka_topic
        self.producer = None
        self.last_status = None
        self.last_exception = None
        self.callback_topic = None
        if producer is None:
            self.setup_kafka_producer()
        else:
            self.producer = producer
        self.message_processor = message_processor

    def setup_kafka_producer(self):
        try:
            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            self.producer = GlobalsSingleton.get().get_kafka_producer()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.eror("Exception occurred while loading schemas {}".format(e))
            raise e

    def get_kafka_topic(self):
        return self.kafka_topic

    def clear_last(self):
        self.last_exception = None
        self.last_status = ResultAvro()

    def get_last_error(self) -> Error:
        return Error(status=self.last_status, e=self.last_exception)

    def get_protocols(self) -> List[ProtocolProxyMng]:
        proto = ProtocolProxyMng()
        proto.set_protocol(protocol=Constants.PROTOCOL_KAFKA)
        proto.set_proxy_class(proxy_class=KafkaProxy.__class__.__name__)
        proto.set_proxy_module(proxy_module=KafkaProxy.__module__)
        result = [proto]
        return result

    def get_type_id(self) -> str:
        raise ManageException("Not implemented")

    def fill_request_by_id_message(self, request: RequestByIdRecord, id_token: str = None, email: str = None,
                                   slice_id: ID = None, slice_name: str = None, states: List[int] = None,
                                   rid: ID = None, delegation_id: str = None, broker_id: ID = None, type: str = None,
                                   site: str = None):
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.callback_topic = self.callback_topic
        request.message_id = str(ID())
        request.id_token = id_token
        request.email = email
        request.states = states
        request.delegation_id = delegation_id
        request.site = site
        request.type = type
        if slice_id is not None:
            request.slice_id = str(slice_id)
        if rid is not None:
            request.reservation_id = str(rid)
        if broker_id is not None:
            request.broker_id = str(broker_id)
        request.slice_name = slice_name

        return request

    def send_request(self, request: AbcMessageAvro) -> Tuple[ResultAvro, Any]:
        self.clear_last()

        status = ResultAvro()
        rret_val = None

        try:
            ret_val = self.producer.produce(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.MANAGEMENT_INTER_ACTOR_OUTBOUND_MESSAGE.format(request.get_message_name(),
                                                                                       self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.MANAGEMENT_API_TIMEOUT_IN_SECONDS)

                if not message_wrapper.done:
                    self.logger.debug(Constants.MANAGEMENT_API_TIMEOUT_OCCURRED)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.interpret()
                else:
                    self.logger.debug(Constants.MANAGEMENT_INTER_ACTOR_INBOUND_MESSAGE.format(message_wrapper.response))
                    status = message_wrapper.response.status
                    if status.code == 0:
                        rret_val = message_wrapper.response
            else:
                self.logger.debug(Constants.MANAGEMENT_INTER_ACTOR_MESSAGE_FAILED.format(
                    request.get_message_name(), self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.interpret()

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.interpret(exception=e)
            status.details = traceback.format_exc()

        self.last_status = status

        return status, rret_val



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import datetime
import json
import logging

from fabric_cm.credmgr.credmgr_proxy import CredmgrProxy, Status
from fabric_cm.credmgr.swagger_client.rest import ApiException
from fss_utils.jwt_manager import ValidateCode
from fss_utils.jwt_validate import JWTValidator

from fabric_cf.actor.core.util.utils import generate_sha256
from fabric_cf.actor.security.fabric_token import TokenException, FabricToken


class TokenValidator:
    """This class caches revoke list retrieved from a specified endpoint
    and uses it to validate provided tokens"""

    def __init__(self, *, credmgr_host: str, refresh_period: datetime.timedelta,
                 jwt_validator: JWTValidator):
        """ Initialize a validator with an endpoint URL presenting Token revoke list,
        a refresh period for keys expressed as datetime.timedelta and
        audience (i.e. CI Logon client id cilogon:/client_id/1234567890).
        :param credmgr_host:
        :param refresh_period:
        """
        self.credmgr_host = credmgr_host
        assert refresh_period is None or isinstance(refresh_period, datetime.timedelta)
        self.cache_period = refresh_period
        self.trl = []
        self.trl_fetched = None
        self.credmgr_proxy = CredmgrProxy(credmgr_host=credmgr_host)
        self.jwt_validator = jwt_validator
        self.logger = logging.getLogger()

    def __fetch_token_revoke_list(self, *, project_id: str):
        """
        Fetch TRL from an endpoint and save it
        @param token token
        """
        if self.trl_fetched is not None:
            if datetime.datetime.now() < self.trl_fetched + self.cache_period:
                return

        status, trl_or_exception = self.credmgr_proxy.token_revoke_list(project_id=project_id)
        if status != Status.OK:
            self.logger.error(f"Exception: {trl_or_exception}")
            details = None
            if isinstance(trl_or_exception, ApiException) and trl_or_exception.body is not None:
                details = json.loads(trl_or_exception.body.decode('utf-8'))
                errors = details.get('errors')
                if errors is not None:
                    details = errors[0].get('details')
            raise Exception(f"Unable to fetch token revoke list: {trl_or_exception.status}:{trl_or_exception.reason}:{details}")

        self.trl_fetched = datetime.datetime.now()
        if trl_or_exception is not None:
            self.trl = trl_or_exception

    def validate_token(self, *, token, verify_exp=False) -> FabricToken:
        """
        Validate a token using a JWT Validator
        Returns the decoded token
        :param token:
        :param verify_exp:
        :return decoded token in a dictionary
        """
        result = None
        token_hash = generate_sha256(token=token)
        if self.jwt_validator is not None:
            code, token_or_exception = self.jwt_validator.validate_jwt(token=token, verify_exp=verify_exp)
            if code is not ValidateCode.VALID:
                raise TokenException(f"Unable to validate provided token: {code}/{token_or_exception}")

            result = FabricToken(decoded_token=token_or_exception, token_hash=token_hash)
            project_id, tags, name = result.first_project
            self.__fetch_token_revoke_list(project_id=project_id)
            if token_hash in self.trl:
                raise TokenException(f"Unauthorized: Token: {token_hash} is revoked!")
        else:
            raise TokenException("JWT Token validator not initialized, skipping validation")

        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Tuple, Dict, List

from fabric_mb.message_bus.messages.poa_avro import PoaAvro
from fabric_mb.message_bus.messages.result_avro import ResultAvro
from fabric_mb.message_bus.messages.result_delegation_avro import ResultDelegationAvro
from fabric_mb.message_bus.messages.result_poa_avro import ResultPoaAvro
from fabric_mb.message_bus.messages.result_reservation_avro import ResultReservationAvro
from fabric_mb.message_bus.messages.result_reservation_state_avro import ResultReservationStateAvro
from fabric_mb.message_bus.messages.result_sites_avro import ResultSitesAvro
from fabric_mb.message_bus.messages.result_slice_avro import ResultSliceAvro
from fabric_mb.message_bus.messages.result_string_avro import ResultStringAvro
from fabric_mb.message_bus.messages.slice_avro import SliceAvro

from fabric_cf.actor.core.apis.abc_management_object import ABCManagementObject
from fabric_cf.actor.core.container.maintenance import Site
from fabric_cf.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.security.auth_token import AuthToken


class ABCActorManagementObject(ABCManagementObject):
    """
    Interface for Management Object for an Actor
    """
    @abstractmethod
    def set_actor(self, *, actor: ABCActorMixin):
        """
        Set an actor
        @params actor: actor
        """

    @abstractmethod
    def update_reservation(self, *, reservation, caller: AuthToken):
        """
        Update Reservation
        @params reservation: reservation
        @params caller: caller
        """

    def get_poas(self, *, caller: AuthToken, states: List[int] = None,
                 slice_id: ID = None, rid: ID = None, email: str = None,
                 poa_id: str = None, project_id: str = None,
                 limit: int = 200, offset: int = 0) -> ResultPoaAvro:
        """
        Get Reservations
        @param states states
        @param slice_id slice ID
        @param rid reservation id
        @param email: user email
        @param poa_id: POA Id
        @param project_id project id
        Obtains all poa with error information in case of failure
        @param caller caller
        @param limit: limit of records to be returned
        @param offset: offset
        @return returns list of the poa
        """

    def is_sliver_provisioning_allowed(self, *, project: str, email: str, site: str,
                                       worker: str) -> Tuple[bool, str or None]:
        """
        Check if sliver provisioning is allowed for a given project on a site/worker
        @param project project id
        @param email user's email
        @param site site
        @param worker worker
        @return True if provisioning is allowed; False otherwise
        """

    def is_slice_provisioning_allowed(self, *, project: str, email: str) -> bool:
        """
        Determine if slice can be provisioned
        Slice provisioning can be prohibited if Testbed is in maintenance mode
        Slice provisioning in maintenance mode may be allowed for specific projects/users
        @param project project
        @param email user's email
        @return True if allowed; False otherwise
        """

    def is_site_in_maintenance(self, *, site_name: str) -> Tuple[bool, Site or None]:
        """
        Check if site is in Maintenance
        @param site_name site name
        @return True, Site Information if site is in Maintenance, False otherwise
        """

    def is_testbed_in_maintenance(self) -> Tuple[bool, Dict[str, str] or None]:
        """
        Check if testbed is in maintenance
        @return True - if testbed is in maintenance, False otherwise
        """

    def remove_delegation(self, *, caller: AuthToken, did: str) -> ResultAvro:
        """
        Removes the specified delegation.
        Note only closed delegation can be removed.
        @param did delegation id of the delegation to be removed
        @param caller caller
        @return result status with error information in case of failure
        """

    @abstractmethod
    def close_delegation(self, *, caller: AuthToken, did: str) -> ResultAvro:
        """
        Closes the specified delegation
        @param did delegation id
        @param caller caller
        @return result status with error information in case of failure
        """

    def get_delegations(self, *, caller: AuthToken, slice_id: ID = None, did: str = None,
                        states: List[int] = None) -> ResultDelegationAvro:
        """
        Get Delegations
        @param slice_id slice id
        @param states states
        @param did delegation id
        @param caller caller
        @return result delegation with error information in case of failure
        """

    def get_reservation_state_for_reservations(self, *, caller: AuthToken,
                                               rids: List[str]) -> ResultReservationStateAvro:
        """
        Returns the state of each of the specified reservations.
        The order in the return list matches the order in the @reservations li
        @param rids list of reservation ids
        @param caller caller
        @return list of state of the specified reservations along with error status in case of failure
        """

    def close_slice_reservations(self, *, caller: AuthToken, slice_id: ID) -> ResultAvro:
        """
        Closes all reservations in the specified slice.
        @param slice_id slice ID
        @param caller caller
        @return result status with error information in case of failure
        """

    def accept_update_slice(self, *, slice_id: ID, caller: AuthToken) -> ResultAvro:
        """
        Accept the last slice update
        @param slice_id slice_id
        @param caller caller
        @return result status with error information in case of failure
        """

    def update_slice(self, *, slice_mng: SliceAvro, caller: AuthToken, modify_state: bool = False) -> ResultAvro:
        """
        Updates the specified slice.
        The only updatable slice attributes are:
        - description
        - all properties lists
        @param slice_mng slice
        @param modify_state - trigger state change
        @param caller caller
        @return result status with error information in case of failure
        """

    def delete_slice(self, *, slice_id: ID, caller: AuthToken) -> ResultAvro:
        """
        Deletes the specified slice - moves the slice into Closing state.
        @param slice_id slice ID
        @param caller caller
        @return result status with error information in case of failure
        """

    def add_slice(self, *, slice_obj: SliceAvro, caller: AuthToken) -> ResultStringAvro:
        """
        Adds a new slice
        @param slice_obj slice
        @param caller caller
        @return returns slice id with error info
        """

    def get_actor(self) -> ABCActorMixin:
        """
        Return Actor
        @return actor
        """

    def close_reservation(self, *, caller: AuthToken, rid: ID) -> ResultAvro:
        """
        Closes the specified reservation
        @param rid reservation id
        @param caller caller
        @return result status with error information in case of failure
        """

    def remove_reservation(self, *, caller: AuthToken, rid: ID) -> ResultAvro:
        """
        Removes the specified reservation.
        Note only closed reservations can be removed.
        @param rid reservation id of the reservation to be removed
        @param caller caller
        @return result status with error information in case of failure
        """

    def get_sites(self, *, caller: AuthToken, site: str) -> ResultSitesAvro:
        """
        @param site site name
        Obtains Maintenance Info for a site
        @param caller caller
        @return returns list of the Sites with error information in case of failure
        """

    def get_reservations(self, *, caller: AuthToken, states: List[int] = None,
                         slice_id: ID = None, rid: ID = None, oidc_claim_sub: str = None,
                         email: str = None, rid_list: List[str] = None, type: str = None,
                         site: str = None, node_id: str = None) -> ResultReservationAvro:
        """
        Get Reservations
        @param states states
        @param slice_id slice ID
        @param rid reservation id
        @param oidc_claim_sub: oidc claim sub
        @param email: user email
        @param rid_list: list of Reservation Id
        @param type type of reservations like NodeSliver/NetworkServiceSliver
        @param site site
        @param node_id node id
        Obtains all reservations with error information in case of failure
        @param caller caller
        @return returns list of the reservations
        """

    def get_slices(self, *, slice_id: ID, caller: AuthToken, slice_name: str = None, email: str = None,
                   states: List[int] = None, project: str = None, limit: int = None,
                   offset: int = None, user_id: str = None) -> ResultSliceAvro:
        """
        Obtains all slices.
        @param slice_id slice id
        @param slice_name slice name
        @param email email
        @param project project id
        @param states slice states
        @param limit limit
        @param offset offset
        @param caller caller
        @param user_id user_id
        @return returns list of slices
        """

    def remove_slice(self, *, slice_id: ID, caller: AuthToken) -> ResultAvro:
        """
        Removes the specified slice
        @param slice_id slice id
        @param caller caller
        @return true for success; false otherwise
        """


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING


from fabric_cf.actor.core.apis.abc_actor_event import ABCActorEvent

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.kernel.retry_rpc import RetryRPC


class RetryRPCEvent(ABCActorEvent):
    """
    Represents Retry RPC Event
    """
    def __init__(self, *, actor: ABCActorMixin, rpc: RetryRPC):
        self.actor = actor
        self.rpc = rpc

    def process(self):
        """
        Process Retry RPC Event
        """
        from fabric_cf.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
        RPCManagerSingleton.get().retry_rpc(rpc=self.rpc)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import List

from fabric_mb.message_bus.messages.get_reservations_request_avro import GetReservationsRequestAvro
from fabric_mb.message_bus.messages.get_reservation_units_request_avro import GetReservationUnitsRequestAvro
from fabric_mb.message_bus.messages.get_unit_request_avro import GetUnitRequestAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.unit_avro import UnitAvro

from fabric_cf.actor.core.apis.abc_mgmt_authority import ABCMgmtAuthority
from fabric_cf.actor.core.apis.abc_reservation_mixin import ReservationCategory
from fabric_cf.actor.core.manage.kafka.kafka_server_actor import KafkaServerActor
from fabric_cf.actor.core.util.id import ID


class KafkaAuthority(KafkaServerActor, ABCMgmtAuthority):
    def get_authority_reservations(self, *, id_token: str = None) -> List[ReservationMng]:
        request = GetReservationsRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token)
        request.reservation_type = ReservationCategory.Authority.name
        status, response = self.send_request(request)

        if status.code == 0:
            return response.reservations
        return None

    def get_reservation_units(self, *, rid: ID, id_token: str = None) -> List[UnitAvro]:
        request = GetReservationUnitsRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token, rid=rid)
        status, response = self.send_request(request)

        if status.code == 0:
            return response.units
        return None

    def get_reservation_unit(self, *, uid: ID, id_token: str = None) -> UnitAvro:
        request = GetUnitRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token, rid=uid)
        status, response = self.send_request(request)

        if status.code == 0:
            return response.units
        return None

    def clone(self):
        return KafkaAuthority(guid=self.management_id,
                              kafka_topic=self.kafka_topic,
                              auth=self.auth, logger=self.logger,
                              message_processor=self.message_processor,
                              producer=self.producer)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric_cf.actor.boot.configuration import ActorConfig
from fabric_cf.actor.core.apis.abc_authority import ABCAuthority
from fabric_cf.actor.core.apis.abc_authority_policy import ABCAuthorityPolicy
from fabric_cf.actor.core.apis.abc_authority_reservation import ABCAuthorityReservation
from fabric_cf.actor.core.apis.abc_broker_reservation import ABCBrokerReservation
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.core.policy import Policy
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.time.term import Term


class AuthorityPolicy(Policy, ABCAuthorityPolicy):
    def __init__(self, *, actor: ABCAuthority = None):
        super().__init__(actor=actor)
        self.initialized = False
        self.delegations = None

    def initialize(self, *, config: ActorConfig):
        if not self.initialized:
            super().initialize(config=config)
            self.initialized = True
            self.delegations = {}

    def revisit(self, *, reservation: ABCReservationMixin):
        return

    def revisit_delegation(self, *, delegation: ABCDelegation):
        self.donate_delegation(delegation=delegation)

    def donate_delegation(self, *, delegation: ABCDelegation):
        self.delegations[delegation.get_delegation_id()] = delegation

    def eject(self, *, resources: ResourceSet):
        return

    def failed(self, *, resources: ResourceSet):
        return

    def freed(self, *, resources: ResourceSet):
        return

    def recovered(self, *, resources: ResourceSet):
        return

    def release(self, *, resources: ResourceSet):
        return

    def bind_delegation(self, *, delegation: ABCDelegation) -> bool:
        result = False

        if delegation.get_delegation_id() not in self.delegations:
            self.delegations[delegation.get_delegation_id()] = delegation
            result = True

        return result

    def allocate(self, *, cycle: int):
        return

    def assign(self, *, cycle: int):
        return

    def correct_deficit(self, *, reservation: ABCAuthorityReservation):
        if reservation.get_resources() is None:
            return

        self.finish_correct_deficit(reservation=reservation)

    def finish_correct_deficit(self, *, reservation: ABCAuthorityReservation, rset: ResourceSet = None):
        """
        Finishes correcting a deficit.
        @param rset correction
        @param reservation reservation
        @raises Exception in case of error
        """
        # We could have a partial set if there's a shortage. Go ahead and
        # install it: we'll come back later for the rest if we return a null
        # term. Alternatively, we could release them and throw an error.
        if rset is None:
            self.log_warn(message="we either do not have resources to satisfy the request or "
                                  "the reservation has/will have a pending operation")
            return

        if rset.is_empty():
            reservation.set_pending_recover(pending_recover=False)
        else:
            reservation.get_resources().update(reservation=reservation, resource_set=rset)

    def extend_authority(self, *, reservation: ABCAuthorityReservation) -> bool:
        return False

    def extend_broker(self, *, reservation: ABCBrokerReservation) -> bool:
        return False

    def extend(self, *, reservation: ABCReservationMixin, resources: ResourceSet, term: Term):
        return False

    def reclaim(self, *, delegation: ABCDelegation):
        return
